from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import base64
import json
import math
import mock
import os
import unittest

from six.moves import cPickle as pickle

from smqtk.algorithms import Classifier
from smqtk.iqr import IqrSession
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.web.classifier_service.classifier_server import \
    SmqtkClassifierService

from .dummy_classifier import DummyClassifier, DummySupervisedClassifier, \
    STUB_CLASSIFIER_MOD_PATH


class TestClassifierService (unittest.TestCase):

    # noinspection PyUnresolvedReferences
    @mock.patch.dict(os.environ, {
        'CLASSIFIER_PATH': STUB_CLASSIFIER_MOD_PATH,
        'DESCRIPTOR_GENERATOR_PATH':
            'smqtk.tests.web.classifier_service.dummy_descriptor_generator'
    })
    def setUp(self):
        super(TestClassifierService, self).setUp()
        self.config = SmqtkClassifierService.get_default_config()

        self.config['classification_factory']['type'] = \
            'MemoryClassificationElement'
        del self.config['classification_factory']['FileClassificationElement']

        del self.config['classifier_collection']['__example_label__']
        self.dummy_label = 'dummy'
        self.config['classifier_collection'][self.dummy_label] = {
            'DummyClassifier': {},
            'type': 'DummyClassifier'
        }
        self.config['immutable_labels'] = [self.dummy_label]

        self.config['descriptor_factory']['type'] = 'DescriptorMemoryElement'
        del self.config['descriptor_factory']['DescriptorFileElement']

        self.config['descriptor_generator'] = {
            'DummyDescriptorGenerator': {},
            'type': 'DummyDescriptorGenerator'
        }

        self.config['iqr_state_classifier_config']['type'] = \
            'DummySupervisedClassifier'

        self.config['enable_classifier_removal'] = True

        self.config['flask_app'] = {}
        del self.config['server']

        self.app = SmqtkClassifierService(json_config=self.config)

    def assertStatus(self, rv, code):
        self.assertEqual(rv.status_code, code)

    def assertResponseMessageRegex(self, rv, regex):
        self.assertRegexpMatches(json.loads(rv.data.decode())['message'], regex)

    def assertMessage(self, resp_data, message):
        self.assertEqual(resp_data['message'], message)

    def test_server_alive(self):
        rv = self.app.test_client().get('/is_ready')

        self.assertStatus(rv, 200)
        resp_data = json.loads(rv.data.decode())
        self.assertMessage(resp_data, "Yes, I'm alive!")

    def test_preconfigured_labels(self):
        rv = self.app.test_client().get('/classifier_labels')

        self.assertStatus(rv, 200)
        resp_data = json.loads(rv.data.decode())
        self.assertMessage(resp_data, "Classifier labels.")
        self.assertListEqual(resp_data['labels'], ['dummy'])

    def test_one_classify(self):
        results_exp = dict(positive=0.5, negative=0.5)
        label = 'dummy'
        content_type = 'text/plain'
        element = base64.b64encode(b'TEST ELEMENT').decode()

        with self.app.test_client() as cli:
            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': element,
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Finished classification.")
            self.assertDictEqual(resp_data['result'][label], results_exp)

            # Same classifier, just retrieving it different ways, so skip the
            # correctness check
            rv = cli.post('/classify', data={
                'label': label,
                'content_type': content_type,
                'bytes_b64': element,
            })
            self.assertStatus(rv, 200)

            rv = cli.post('/classify', data={
                'label': json.dumps(label),
                'content_type': content_type,
                'bytes_b64': element,
            })
            self.assertStatus(rv, 200)

            rv = cli.post('/classify', data={
                'label': json.dumps([label]),
                'content_type': content_type,
                'bytes_b64': element,
            })
            self.assertStatus(rv, 200)

    def test_adjusted_classify(self):
        results_exp = dict(positive=0.5, negative=0.5)
        label = 'dummy'
        content_type = 'text/plain'
        element = base64.b64encode(b'TEST ELEMENT').decode()

        with self.app.test_client() as cli:
            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': element,
                'adjustment': json.dumps({
                    'positive': 0,
                }),
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Finished classification.")
            self.assertDictEqual(resp_data['result'][label], results_exp)

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': element,
                'adjustment': json.dumps({
                    'positive': -1,
                }),
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Finished classification.")
            result = resp_data['result'][label]
            self.assertAlmostEqual(result['positive'], 1/(1+math.exp(-1)))
            self.assertAlmostEqual(result['negative'], 1/(1+math.exp(1)))

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': element,
                'adjustment': json.dumps({
                    'positive': 1,
                }),
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Finished classification.")
            result = resp_data['result'][label]
            self.assertAlmostEqual(result['positive'], 1/(1+math.exp(1)))
            self.assertAlmostEqual(result['negative'], 1/(1+math.exp(-1)))

    def test_multiple_classify(self):
        content_type = 'text/plain'
        element = base64.b64encode(b'TEST ELEMENT').decode()
        results_exp = dict(positive=0.5, negative=0.5)
        pickle_data = pickle.dumps(DummyClassifier.from_config({}))
        enc_data = base64.b64encode(pickle_data)
        old_label = 'dummy'
        new_label = 'dummy2'
        lock_clfr_str = 'true'

        with self.app.test_client() as cli:
            rv = cli.post('/classifier', data={
                'label': new_label,
                'lock_label': lock_clfr_str,
                'bytes_b64': enc_data,
            })
            self.assertStatus(rv, 201)
            resp_data = json.loads(rv.data.decode())
            self.assertEqual(resp_data["message"],
                             "Uploaded classifier for label '%s'."
                             % new_label)
            self.assertEqual(resp_data["label"], new_label)

            rv = cli.post('/classify', data={
                'label': new_label,
                'content_type': content_type,
                'bytes_b64': element,
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Finished classification.")
            self.assertDictEqual(resp_data['result'][new_label], results_exp)

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': element,
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Finished classification.")
            self.assertDictEqual(resp_data['result'][old_label], results_exp)
            self.assertDictEqual(resp_data['result'][new_label], results_exp)

    def test_get_add_del_classifier(self):
        old_label = 'dummy'
        new_label = 'dummy2'

        with self.app.test_client() as cli:
            rv = cli.get('/classifier', data={
                'label': old_label,
            })
            self.assertStatus(rv, 200)
            enc_data = rv.data.decode()

            rv = cli.post('/classifier', data={
                'label': new_label,
                'bytes_b64': enc_data,
            })
            self.assertStatus(rv, 201)
            resp_data = json.loads(rv.data.decode())
            self.assertEqual(resp_data["message"],
                             "Uploaded classifier for label '%s'."
                             % new_label)
            self.assertEqual(resp_data["label"], new_label)

            rv = cli.get('/classifier_labels')
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Classifier labels.")
            self.assertSetEqual(set(resp_data['labels']),
                                {old_label, new_label})

            rv = cli.delete('/classifier', data={
                'label': new_label,
            })
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertEqual(resp_data["message"],
                             "Removed classifier with label '%s'."
                             % new_label)
            self.assertEqual(resp_data["removed_label"], new_label)

            rv = cli.get('/classifier_labels')
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Classifier labels.")
            self.assertSetEqual(set(resp_data['labels']), {old_label})

    def test_add_imm_del_classifier(self):
        pickle_data = pickle.dumps(DummyClassifier.from_config({}))
        enc_data = base64.b64encode(pickle_data).decode('utf8')
        old_label = 'dummy'
        new_label = 'dummy2'
        lock_clfr_str = 'true'

        with self.app.test_client() as cli:
            rv = cli.post('/classifier', data={
                'label': new_label,
                'lock_label': lock_clfr_str,
                'bytes_b64': enc_data,
            })
            self.assertStatus(rv, 201)
            resp_data = json.loads(rv.data.decode())
            self.assertEqual(resp_data["message"],
                             "Uploaded classifier for label '%s'."
                             % new_label)
            self.assertEqual(resp_data["label"], new_label)

            rv = cli.get('/classifier_labels')
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Classifier labels.")
            self.assertSetEqual(set(resp_data['labels']),
                                {old_label, new_label})

            rv = cli.delete('/classifier', data={
                'label': new_label,
            })
            self.assertStatus(rv, 405)
            resp_data = json.loads(rv.data.decode())
            self.assertEqual(resp_data["message"],
                             "Label '%s' refers to a classifier that is"
                             " immutable." % new_label)
            self.assertEqual(resp_data['label'], new_label)

            rv = cli.get('/classifier_labels')
            self.assertStatus(rv, 200)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data, "Classifier labels.")
            self.assertSetEqual(set(resp_data['labels']),
                                {old_label, new_label})

    def test_post_classifier_failures(self):
        pickle_data = pickle.dumps(DummyClassifier.from_config({}))
        enc_data = base64.b64encode(pickle_data)
        bad_data = base64.b64encode(pickle.dumps(object()))
        old_label = 'dummy'
        new_label = 'dummy2'
        lock_clfr_str = '['

        with self.app.test_client() as cli:
            rv = cli.post('/classifier', data={
                'label': old_label,
                'bytes_b64': enc_data,
            })
            self.assertStatus(rv, 400)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data,
                               "Label '%s' already exists in classifier"
                               " collection." % old_label)
            self.assertEqual(resp_data['label'], old_label)

            rv = cli.post('/classifier', data={'label': old_label})
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "No state base64 data provided.")

            rv = cli.post('/classifier', data={'bytes_b64': enc_data})
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "No descriptive label provided.")

            rv = cli.post('/classifier', data={
                'label': old_label,
                'lock_label': lock_clfr_str,
                'bytes_b64': enc_data,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "Invalid boolean value for 'lock_label'."
                               " Was given: '%s'" % lock_clfr_str)

            rv = cli.post('/classifier', data={
                'label': new_label,
                'bytes_b64': bad_data,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "Data added for label '%s' is not a"
                               " Classifier." % new_label)

    def test_del_classifier_failures(self):
        old_label = 'dummy'
        new_label = 'dummy2'

        with self.app.test_client() as cli:
            rv = cli.delete('/classifier', data={})
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()), "No label provided.")

            rv = cli.delete('/classifier', data={'label': old_label})
            self.assertStatus(rv, 405)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data,
                               "Label '%s' refers to a classifier that is"
                               " immutable." % old_label)
            self.assertEqual(resp_data['label'], old_label)

            rv = cli.delete('/classifier', data={'label': new_label})
            self.assertStatus(rv, 404)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data,
                               "Label '%s' does not refer to a classifier"
                               " currently registered." % new_label)
            self.assertEqual(resp_data['label'], new_label)

    def test_get_classifier_failures(self):
        label = 'dummy2'

        with self.app.test_client() as cli:
            rv = cli.get('/classifier', data={})
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()), "No label provided.")

            rv = cli.get('/classifier', data={'label': label})
            self.assertStatus(rv, 404)
            resp_data = json.loads(rv.data.decode())
            self.assertMessage(resp_data,
                               "Label '%s' does not refer to a classifier"
                               " currently registered." % label)
            self.assertEqual(resp_data['label'], label)

    def test_classify_failures(self):
        content_type = 'text/plain'
        bytes_b64 = base64.b64encode(b'TEST ELEMENT').decode()
        label_invalid_json_failure = '['
        label_valid_json_failure = '{}'
        label_valid_json_list_failure = '["test", {}]'
        missing_clfrs_1 = ['dummy', 'foo']
        missing_clfrs_2 = ['dummy', 'foo', 'bar']
        missing_clfrs_3 = ['foo']
        missing_clfrs_4 = ['foo', 'bar']

        with self.app.test_client() as cli:
            rv = cli.post('/classify', data={
                'content_type': content_type,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "No base-64 bytes provided.")

            rv = cli.post('/classify', data={
                'bytes_b64': bytes_b64,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "No content type provided.")

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': label_invalid_json_failure,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "Label(s) are not properly formatted JSON.")

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': label_valid_json_failure,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "Label must be a list of strings or a single"
                               " string.")

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': label_valid_json_list_failure,
            })
            self.assertStatus(rv, 400)
            self.assertMessage(json.loads(rv.data.decode()),
                               "Label must be a list of strings or a single"
                               " string.")

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': json.dumps(missing_clfrs_1),
            })
            self.assertStatus(rv, 404)
            resp_data = json.loads(rv.data.decode())
            self.assert_(resp_data['message'].startswith(
                "The following labels are not registered with any"
                " classifiers:"))
            self.assertSetEqual(set(resp_data['missing_labels']),
                                set(missing_clfrs_1) - {'dummy'})

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': json.dumps(missing_clfrs_2),
            })
            self.assertStatus(rv, 404)
            resp_data = json.loads(rv.data.decode())
            self.assert_(resp_data['message'].startswith(
                "The following labels are not registered with any"
                " classifiers:"))
            self.assertSetEqual(set(resp_data['missing_labels']),
                                set(missing_clfrs_2) - {'dummy'})

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': json.dumps(missing_clfrs_3),
            })
            self.assertStatus(rv, 404)
            resp_data = json.loads(rv.data.decode())
            self.assert_(resp_data['message'].startswith(
                "The following labels are not registered with any"
                " classifiers:"))
            self.assertSetEqual(set(resp_data['missing_labels']),
                                set(missing_clfrs_3))

            rv = cli.post('/classify', data={
                'content_type': content_type,
                'bytes_b64': bytes_b64,
                'label': json.dumps(missing_clfrs_4),
            })
            self.assertStatus(rv, 404)
            resp_data = json.loads(rv.data.decode())
            self.assert_(resp_data['message'].startswith(
                "The following labels are not registered with any"
                " classifiers:"))
            self.assertSetEqual(set(resp_data['missing_labels']),
                                set(missing_clfrs_4))

    def test_get_classifier_metadata_no_label(self):
        with self.app.test_client() as cli:
            #: :type: flask.wrappers.Response
            r = cli.get('/classifier_metadata')
            r_json = json.loads(r.data.decode())
            self.assertStatus(r, 400)
            self.assertMessage(r_json, "No label provided.")

    def test_get_classifier_metadata_invalid_label(self):
        with self.app.test_client() as cli:
            args = dict(label="no-valid-label")
            r = cli.get('/classifier_metadata', query_string=args)
            r_json = json.loads(r.data.decode())
            self.assertStatus(r, 404)
            self.assertMessage(r_json, "Label 'no-valid-label' does not refer "
                                       "to a classifier currently registered.")

    def test_get_classifier_labels_mocked(self):
        """
        Test that we can request the registered dummy classifiers class labels.
        Using mock objects to assert calls made.
        """
        expected_label = 'this-test-label'
        expected_class_labels = ['foo', 'bar', 'shazam']

        mock_classifier = mock.Mock(spec=Classifier)
        mock_classifier.get_labels = \
            mock.Mock(return_value=expected_class_labels)

        self.app.classifier_collection.labels = mock.Mock(
            return_value={expected_label}
        )
        self.app.classifier_collection.get_classifier = mock.Mock(
            return_value=mock_classifier
        )

        with self.app.test_client() as cli:
            args = dict(label=expected_label)
            r = cli.get('/classifier_metadata', query_string=args)

            self.app.classifier_collection.labels.assert_called_once_with()
            self.app.classifier_collection.get_classifier\
                    .assert_called_once_with(expected_label)
            mock_classifier.get_labels.assert_called_once_with()

            r_json = json.loads(r.data.decode())
            self.assertStatus(r, 200)
            self.assertMessage(r_json, "Success")
            self.assertIn('class_labels', r_json)
            self.assertListEqual(r_json['class_labels'], expected_class_labels)

    def test_get_classifier_labels(self):
        """
        Test that we can request the registered dummy classifiers class labels.
        """
        with self.app.test_client() as cli:
            args = dict(label=self.dummy_label)
            r = cli.get('/classifier_metadata', query_string=args)
            r_json = json.loads(r.data.decode())
            self.assertStatus(r, 200)
            self.assertMessage(r_json, "Success")
            self.assertIn('class_labels', r_json)
            self.assertSetEqual(set(r_json['class_labels']),
                                {'negative', 'positive'})

    def test_add_iqr_state_classifier_param_failures(self):
        test_bytes = b"some not used bytes"
        test_bytes_b64 = base64.b64encode(test_bytes).decode()
        test_label = "classifier-test-label"

        with self.app.test_client() as cli:
            # Missing Bytes
            rv = cli.post('/iqr_classifier')
            self.assertStatus(rv, 400)
            self.assertResponseMessageRegex(
                rv, "No state base64 data provided."
            )

            # Missing label for classifier
            rv = cli.post('/iqr_classifier', data={
                "bytes_b64": test_bytes_b64,
            })
            self.assertStatus(rv, 400)
            self.assertResponseMessageRegex(
                rv, "No descriptive label provided."
            )

            # Invalid lock flag value (not a boolean)
            rv = cli.post('/iqr_classifier', data={
                'bytes_b64': test_bytes_b64,
                'label': test_label,
                'lock_label': 'not-bool-convertible'
            })
            self.assertStatus(rv, 400)
            self.assertResponseMessageRegex(
                rv, "Invalid boolean value for 'lock_label'. Was given: "
            )

    def test_add_iqr_state_classifier_existing_label(self):
        test_label = 'duplicate-label'
        test_new_cfier_b64 = base64.b64encode(
            b"some not used bytes"
        )

        self.app.classifier_collection.add_classifier(
            test_label, DummySupervisedClassifier()
        )

        with self.app.test_client() as cli:
            rv = cli.post('/iqr_classifier', data={
                'bytes_b64': test_new_cfier_b64,
                'label': test_label,
            })
            self.assertStatus(rv, 400)
            self.assertResponseMessageRegex(
                rv, "Label already exists in classifier collection."
            )

    @mock.patch.dict(os.environ, {"CLASSIFIER_PATH": STUB_CLASSIFIER_MOD_PATH})
    def test_add_iqr_state_classifier_simple(self):
        """
        Test calling IQR classifier add endpoint with a simple IQR Session
        serialization.
        """
        # Make a simple session with dummy adjudication descriptor elements
        iqrs = IqrSession(session_uid=str("0"))
        iqr_p1 = DescriptorMemoryElement('test', 0).set_vector([0])
        iqr_n1 = DescriptorMemoryElement('test', 1).set_vector([1])
        iqrs.adjudicate(
            new_positives=[iqr_p1], new_negatives=[iqr_n1]
        )

        test_iqrs_b64 = base64.b64encode(iqrs.get_state_bytes())
        test_label = 'test-label-08976azsdv'

        with mock.patch(STUB_CLASSIFIER_MOD_PATH +
                        ".DummySupervisedClassifier._train") as m_cfier_train:

            with self.app.test_client() as cli:
                rv = cli.post('/iqr_classifier', data={
                    'bytes_b64': test_iqrs_b64,
                    'label': test_label,
                })
                self.assertStatus(rv, 201)
                self.assertResponseMessageRegex(rv, "Finished training "
                                                    "IQR-session-based "
                                                    "classifier for label "
                                                    "'%s'." % test_label)

            m_cfier_train.assert_called_once_with(
                {'positive': {iqr_p1}, 'negative': {iqr_n1}}
            )
            # Collection should include initial dummy classifier and new iqr
            # classifier.
            self.assertEqual(len(self.app.classifier_collection.labels()), 2)
            self.assertIn(test_label, self.app.classifier_collection.labels())



from __future__ import division, print_function

import pickle
import unittest

import six

from smqtk.exceptions import ReadOnlyError
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.data_set.memory_set import DataMemorySet


class TestDataFileSet (unittest.TestCase):

    def test_is_usable(self):
        # no dependencies
        self.assertTrue(DataMemorySet.is_usable())

    def test_default_config(self):
        default_config = DataMemorySet.get_default_config()
        self.assertEqual(len(default_config), 2)
        self.assertIn('cache_element', default_config)
        self.assertIsInstance(default_config['cache_element'],
                              dict)
        self.assertIsNone(default_config['cache_element']['type'])
        self.assertIn('pickle_protocol', default_config)

    def test_from_config_default(self):
        # From default configuration, which should be valid. Specifies no cache
        # pickle protocol -1.
        c = DataMemorySet.get_default_config()
        i = DataMemorySet.from_config(c)
        self.assertIsNone(i.cache_element)
        self.assertEqual(i.pickle_protocol, -1)
        self.assertEqual(i._element_map, {})

    def test_from_config_empty_cache(self):
        # Specify a memory element cache with no pre-existing bytes.
        c = DataMemorySet.get_default_config()
        c['cache_element']['type'] = 'DataMemoryElement'
        i = DataMemorySet.from_config(c)
        self.assertIsNotNone(i.cache_element)
        self.assertIsInstance(i.cache_element, DataMemoryElement)
        self.assertEqual(i.cache_element.get_bytes(), six.b(''))
        self.assertEqual(i.pickle_protocol, -1)
        self.assertEqual(i._element_map, {})

    def test_from_config_with_cache(self):
        # Use a cache element with bytes defining pickle of map to use.
        expected_map = dict(a=1, b=2, c=3)

        c = DataMemorySet.get_default_config()
        c['cache_element']['type'] = 'DataMemoryElement'
        c['cache_element']['DataMemoryElement']['bytes'] = \
            pickle.dumps(expected_map)

        i = DataMemorySet.from_config(c)

        self.assertIsInstance(i.cache_element, DataMemoryElement)
        self.assertEqual(i.pickle_protocol, -1)
        self.assertEqual(i._element_map, expected_map)

    def test_init_no_cache(self):
        i = DataMemorySet()
        self.assertIsNone(i.cache_element)
        self.assertEqual(i._element_map, {})
        self.assertEqual(i.pickle_protocol, -1)

    def test_init_empty_cache(self):
        cache_elem = DataMemoryElement()
        i = DataMemorySet(cache_elem, 2)
        self.assertEqual(i.cache_element, cache_elem)
        self.assertEqual(i.pickle_protocol, 2)
        self.assertEqual(i._element_map, {})

    def test_init_with_cache(self):
        expected_map = dict(a=1, b=2, c=3)
        expected_cache = DataMemoryElement(bytes=pickle.dumps(expected_map))

        i = DataMemorySet(expected_cache)

        self.assertEqual(i.cache_element, expected_cache)
        self.assertEqual(i.pickle_protocol, -1)
        self.assertEqual(i._element_map, expected_map)

    def test_iter(self):
        expected_map = {
            0: 'a',
            75: 'b',
            124769: 'c',
        }
        expected_map_values = {'a', 'b', 'c'}

        dms = DataMemorySet()
        dms._element_map = expected_map
        self.assertEqual(set(dms), expected_map_values)
        self.assertEqual(set(iter(dms)), expected_map_values)

    def test_caching_no_map_no_cache(self):
        dms = DataMemorySet()
        # should do nothing
        dms.cache()
        self.assertIsNone(dms.cache_element)
        self.assertEqual(dms._element_map, {})

    def test_cacheing_no_map(self):
        dms = DataMemorySet(DataMemoryElement())
        dms.cache()
        # technically caches something, but that something is an empty map.
        self.assertFalse(dms.cache_element.is_empty())
        self.assertEqual(pickle.loads(dms.cache_element.get_bytes()), {})

    def test_cacheing_with_map(self):
        expected_cache = DataMemoryElement()
        expected_map = {
            0: 'a',
            75: 'b',
            124769: 'c',
        }

        dms = DataMemorySet(expected_cache)
        dms._element_map = expected_map
        dms.cache()

        self.assertFalse(expected_cache.is_empty())
        self.assertEqual(pickle.loads(expected_cache.get_bytes()), expected_map)

    def test_caching_readonly_cache(self):
        ro_cache = DataMemoryElement(readonly=True)
        dms = DataMemorySet(ro_cache)
        self.assertRaises(
            ReadOnlyError,
            dms.cache
        )

    def test_get_config_from_config_idempotence(self):
        default_c = DataMemorySet.get_default_config()
        self.assertEqual(
            DataMemorySet.from_config(default_c).get_config(),
            default_c
        )

        c = DataMemorySet.get_default_config()
        c['cache_element']['type'] = 'DataMemoryElement'
        c['cache_element']['DataMemoryElement']['readonly'] = True
        c['pickle_protocol'] = 1
        self.assertEqual(
            DataMemorySet.from_config(c).get_config(),
            c
        )

    def test_count(self):
        expected_map = {
            0: 'a',
            75: 'b',
            124769: 'c',
        }

        dms = DataMemorySet()
        dms._element_map = expected_map
        self.assertEqual(dms.count(), 3)

    def test_uuids(self):
        expected_map = {
            0: 'a',
            75: 'b',
            124769: 'c',
        }

        dms = DataMemorySet()
        dms._element_map = expected_map
        self.assertEqual(dms.uuids(), {0, 75, 124769})

    def test_has_uuid(self):
        expected_map = {
            0: 'a',
            75: 'b',
            124769: 'c',
        }

        dms = DataMemorySet()
        dms._element_map = expected_map
        self.assertTrue(dms.has_uuid(0))
        self.assertTrue(dms.has_uuid(75))
        self.assertTrue(dms.has_uuid(124769))

    def test_add_data_not_DataElement(self):
        dms = DataMemorySet()
        self.assertRaises(
            AssertionError,
            dms.add_data, "not data element"
        )

    def test_add_data(self):
        de = DataMemoryElement(six.b('some bytes'), 'text/plain', True)
        expected_map = {de.uuid(): de}

        dms = DataMemorySet()
        dms.add_data(de)
        self.assertEqual(dms._element_map, expected_map)

    def test_get_data_invalid_uuid(self):
        dms = DataMemorySet()
        self.assertRaises(
            KeyError,
            dms.get_data, 'invalid uuid'
        )

    def test_get_data_valid_uuid(self):
        expected_map = {
            0: 'a',
            75: 'b',
            124769: 'c',
        }

        dms = DataMemorySet()
        dms._element_map = expected_map
        self.assertEqual(dms.get_data(0), 'a')
        self.assertEqual(dms.get_data(75), 'b')
        self.assertEqual(dms.get_data(124769), 'c')



import random

# noinspection PyUnresolvedReferences
from six.moves import range


def partition_string(s, segments):
    """
    Partition a string into a number of segments. If the given number of
    segments does not divide evenly into the string's length, extra characters
    are added to the leading segments in order to allow for the requested number
    of segments.

    This is useful when partitioning the checksum of a file to determine where
    it should be placed in a directory tree system.

    >>> partition_string("foobar", 2)
    ['foo', 'bar']
    >>> partition_string("foobar", 4)
    ['fo', 'ob', 'a', 'r']
    >>> partition_string("foobar", 6)
    ['f', 'o', 'o', 'b', 'a', 'r']

    If the string is not evenly divisible by the requested number of segments,
    then the length of trailing segments will be shorter than leading segments.

    >>> partition_string('d7ca25c5-b886-4a1b-87fe-5945313d350b', 11)
    ['d7ca', '25c5', '-b88', '6-4', 'a1b', '-87', 'fe-', '594', '531', '3d3',
     '50b']
    >>> partition_string('abcde', 2)
    ['abc', 'de']
    >>> partition_string('abcde', 4)
    ['ab', 'c', 'd', 'e']

    If the number of segments is greater than the number of characters in the
    input string, an assertion error is raised.

    >>> partition_string('a', 2)
    Traceback (most recent call last):
        ...
    AssertionError: Cannot split given string into more segments than there are
    characters in the string!

    :raises AssertionError: Segmentation value greater than the length of the
        string.

    :param s: String to partition.
    :type s: str

    :param segments: Number of segments to split the string into
    :type segments: int

    :return: A list of N segments. If the given number of segments does not
        divide evenly into the string's length, this function behaves
        erratically.
    :rtype: list[str]

    """
    assert segments <= len(s), \
        "Cannot split given string into more segments than there are " \
        "characters in the string!"

    seg_len = len(s) // segments
    extra_iters = len(s) % segments

    r = []
    i = 0
    for k in range(segments):
        j = i + seg_len + (k < extra_iters)
        r.append(s[i:j])
        i = j

    return r


DEFAULT_CHAR_SET = \
    'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'


def random_characters(n, char_set=DEFAULT_CHAR_SET):
    """
    Return ``n`` random characters from the given ``char_set``.

    If ``n`` is a floating point valid, it is cast to an integer (floor).
    The default ``char_set`` includes a-z, A-Z and 0-9.

    :param n: Number of random characters to return.
    :type n: int

    :param char_set: Sequence of characters to pull from when constructing
        random sequence.
    :type char_set: str | unicode

    :return: New string of random characters of length ``n`` from the given
        ``char_set``.
    :rtype: str | unicode

    :raises ValueError: If ``char_set`` given is empty, or ``n`` is negative.

    """
    n = int(n)
    if n < 0:
        raise ValueError("n must be a positive integer.")
    L = len(char_set)
    if L == 0:
        raise ValueError("Empty char_set given.")
    return ''.join(char_set[random.randint(0, L - 1)] for _ in range(n))



import threading

import numpy as np
from six import BytesIO, next
from six.moves import map
from sklearn.neighbors import BallTree, DistanceMetric

from smqtk.algorithms.nn_index.hash_index import HashIndex
from smqtk.representation import get_data_element_impls
from smqtk.utils import merge_dict, plugin


class SkLearnBallTreeHashIndex (HashIndex):
    """
    Hash index using the ball tree implementation in scikit-learn.

    *Note:* **When saving this object's model or pickling, we do not naively
    pickle the underlying ball tree due to issues when saving the state of a
    large ball tree. We instead get the state and split its contents up for
    separate serialization via known safe methods.**
    """

    @classmethod
    def is_usable(cls):
        return BallTree is not None

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as
        arguments, turning those argument names into configuration dictionary
        keys. If any of those arguments have defaults, we will add those values
        into the configuration dictionary appropriately. The dictionary
        returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this
        class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(SkLearnBallTreeHashIndex, cls).get_default_config()
        c['cache_element'] = plugin.make_config(get_data_element_impls())
        return c

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless an instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: SkLearnBallTreeHashIndex

        """
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        # Parse ``cache_element`` configuration if set.
        cache_element = None
        if config_dict['cache_element'] and \
                config_dict['cache_element']['type']:
            cache_element = \
                plugin.from_plugin_config(config_dict['cache_element'],
                                          get_data_element_impls())
        config_dict['cache_element'] = cache_element

        return super(SkLearnBallTreeHashIndex, cls).from_config(config_dict,
                                                                False)

    def __init__(self, cache_element=None, leaf_size=40, random_seed=None):
        """
        Initialize Scikit-Learn BallTree index for hash codes.

        :param cache_element: Optional data element to cache our index to.
        :type cache_element: smqtk.representation.DataElement | None

        :param leaf_size: Number of points at which to switch to brute-force.
        :type leaf_size: int

        :param random_seed: Optional random number generator seed (numpy).
        :type random_seed: None | int

        """
        super(SkLearnBallTreeHashIndex, self).__init__()
        self.cache_element = cache_element
        self.leaf_size = leaf_size
        self.random_seed = random_seed

        self._model_lock = threading.RLock()

        # the actual index
        #: :type: sklearn.neighbors.BallTree
        self.bt = None

        self.load_model()

    def get_config(self):
        c = merge_dict(self.get_default_config(), {
            'leaf_size': self.leaf_size,
            'random_seed': self.random_seed,
        })
        if self.cache_element:
            c['cache_element'] = merge_dict(c['cache_element'],
                                            plugin.to_plugin_config(
                                                self.cache_element))
        return c

    def save_model(self):
        """
        Cache a built B-Tree index to the configured cache element. This only
        occurs if we have a non-null cache element and a btree to save.

        :raises ValueError: If the cache element configured is not writable.

        """
        with self._model_lock:
            if self.cache_element:
                if self.cache_element.is_read_only():
                    raise ValueError("Configured cache element (%s) is read-"
                                     "only." % self.cache_element)
                if self.bt is not None:
                    self._log.debug("Saving model: %s", self.cache_element)
                    # Saving BT component matrices separately.
                    # - Not saving distance function because its always going
                    #   to be hamming distance (see ``build_index``).
                    s = self.bt.__getstate__()
                    # Everything after known numpy ndarrays we consider as a
                    # chunk.  Importantly, index 11 should be the distance
                    # metric function used in the model.  We currently assume
                    # its a hamming distance metric.
                    tail = s[4:]
                    buff = BytesIO()
                    # noinspection PyTypeChecker
                    np.savez(buff,
                             data_arr=s[0],
                             idx_array_arr=s[1],
                             node_data_arr=s[2],
                             node_bounds_arr=s[3],
                             tail=tail)
                    self.cache_element.set_bytes(buff.getvalue())
                    self._log.debug("Saving model: Done")
                else:
                    # No ball tree, empty cache.
                    self.cache_element.set_bytes('')

    def load_model(self):
        """
        Load a btree index from the configured cache element. This only occurs
        if there is a cache element configured and there are bytes there to
        read.
        """
        with self._model_lock:
            if self.cache_element and not self.cache_element.is_empty():
                self._log.debug("Loading model from cache: %s",
                                self.cache_element)
                buff = BytesIO(self.cache_element.get_bytes())
                with np.load(buff) as cache:
                    tail = tuple(cache['tail'])
                    s = [cache['data_arr'], cache['idx_array_arr'],
                         cache['node_data_arr'], cache['node_bounds_arr']]
                    s.extend(tail)
                    s[11] = DistanceMetric.get_metric('hamming')
                    s = tuple(s)
                # noinspection PyTypeChecker
                #: :type: sklearn.neighbors.BallTree
                self.bt = BallTree.__new__(BallTree)
                self.bt.__setstate__(s)
                self._log.debug("Loading mode: Done")

    def count(self):
        with self._model_lock:
            return self.bt.data.shape[0] if self.bt else 0

    def _build_bt_internal(self, vec_list):
        """
        Internal shared BallTree build function given a list of boolean hash
        vectors.

        This is here to avoid calling `_build_index` from `_update...` and
        `_remove...` which cuts down on redundant data consistency checking.

        :param vec_list: List of bool-valued numpy arrays.
        :type vec_list: list[np.ndarray[bool]] | np.ndarray[np.ndarray[bool]]

        """
        if len(vec_list) > 0:
            # If distance metric ever changes, need to update save/load model
            # functions.
            self.bt = BallTree(vec_list, self.leaf_size,
                               metric='hamming')
        else:
            self.bt = None
        self.save_model()

    def _build_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to build the index
        with the given hash codes (bit-vectors).

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :param hashes: Iterable of descriptor elements to build index
            over.
        :type hashes: collections.Iterable[np.ndarray[bool]]
        """
        with self._model_lock:
            self._log.debug("Building ball tree")
            if self.random_seed is not None:
                np.random.seed(self.random_seed)
            # BallTree can't take iterables, so catching input in a set of
            # tuples first in order to cull out duplicates (BT will index
            # duplicate values happily).
            hash_tuple_set = set(map(lambda v: tuple(v), hashes))
            # Convert tuples back into numpy arrays for BallTree constructor.
            hash_vector_list = list(map(lambda t: np.array(t), hash_tuple_set))
            self._build_bt_internal(hash_vector_list)

    def _update_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more hash vectors given.

        If no index exists yet, a new one should be created using the given
        hash vectors.

        *Note:* The scikit-learn ball-tree implementation does not support
        incremental updating of its model, so we need to rebuild the model from
        scratch using the currently indexed hashes and the new ones provided to
        this method.

        :param hashes: Iterable of numpy boolean hash vectors to add to this
            index.
        :type hashes: collections.Iterable[np.ndarray[bool]]

        """
        with self._model_lock:
            # Can't use iterators with numpy operations.
            new_hashes = tuple(hashes)
            if self.bt is None:
                # 0-row array using bit-vector size of first new entry length.
                # - Must have at least one new hash due to super-method check.
                indexed_hash_vectors = np.ndarray((0, len(new_hashes[0])))
            else:
                indexed_hash_vectors = self.bt.data
            # Build a new index as normal with the union of source data.
            self._log.debug("Updating index by rebuilding with union.")
            self._build_bt_internal(
                np.concatenate([indexed_hash_vectors, new_hashes], 0)
            )

    def _remove_from_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to partially remove
        hashes from this index.

        *Note:* The scikit-learn ball-tree implementation does not support
        incremental removal from its model, so we need to rebuild the model
        from scratch using the currently indexed hashes minus the ones provided
        to this method.

        :param hashes: Iterable of numpy boolean hash vectors to remove from
            this index.
        :type hashes: collections.Iterable[np.ndarray[bool]]

        :raises KeyError: One or more hashes provided do not match any stored
            hashes.  The index should not be modified.

        """
        with self._model_lock:
            # Convert to a set of hashable tuples for removal
            if self.bt:
                tuple_matrix = set(tuple(r) for r in
                                   np.asarray(self.bt.data).tolist())
                hash_tuples = set()
                for h in hashes:
                    # this is oddly faster than tuple(h)
                    h_t = tuple(h.tolist())
                    if h_t not in tuple_matrix:
                        raise KeyError(h)
                    hash_tuples.add(h_t)
                new_data = tuple_matrix - hash_tuples
                self._build_bt_internal(
                    list(map(lambda t: np.array(t), new_data))
                )
            else:
                # No index built, so anything is a key error.
                # We can also only be here if hashes was non-zero in size.
                raise KeyError(np.asarray(next(iter(hashes))))

    def _nn(self, h, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbor hash codes as bit-vectors to the given hash code
        bit-vector.

        Distances are in the range [0,1] and are the percent different each
        neighbor hash is from the query, based on the number of bits contained
        in the query (normalized hamming distance).

        When this internal method is called, we have already checked that our
        index is not empty.

        :param h: Hash code to compute the neighbors of. Should be the same bit
            length as indexed hash codes.
        :type h: np.ndarray[bool]

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N hash codes and a tuple of the distance
            values to those neighbors.
        :rtype: (tuple[np.ndarray[bool]], tuple[float])

        """
        with self._model_lock:
            # Reselect N based on how many hashes are currently indexes
            n = min(n, self.count())
            # Reshaping ``h`` into an array of arrays, with just one array
            # (ball tree deprecation warns when giving it a single array).
            dists, idxs = self.bt.query([h], n, return_distance=True)
            # only indexing the first entry became we're only querying with one
            # vector
            neighbors = np.asarray(self.bt.data)[idxs[0]].astype(bool)
            return neighbors, dists[0]



import time

import numpy

from smqtk.representation import DescriptorElement


# Try to import required module
try:
    import solr
except ImportError:
    solr = None


class SolrDescriptorElement (DescriptorElement):
    """
    Descriptor element that uses a Solr instance as the backend storage medium.

    Fields where data is stored in the Solr documents are specified at
    construction time. We additionally set the ``id`` field to a string UUID.
    ``id`` is set because it is a common, required field for unique
    identification of documents. The value set to the ``id`` field is
    reproducible from this object's key attributes.

    """

    @classmethod
    def is_usable(cls):
        return solr is not None

    def __init__(self, type_str, uuid, solr_conn_addr,
                 type_field, uuid_field, vector_field, timestamp_field,
                 timeout=10, persistent_connection=False, commit_on_set=True):
        """
        Initialize a new Solr-stored descriptor element.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        :param solr_conn_addr: HTTP(S) address for the Solr index to use
        :type solr_conn_addr: str

        :param type_field: Solr index field to store descriptor type string
            value.
        :type type_field: str

        :param uuid_field: Solr index field to store descriptor UUID string
            value in.
        :type uuid_field: str

        :param vector_field: Solr index field to store the descriptor vector of
            floats in.
        :type vector_field: str

        :param timestamp_field: Solr index field to store floating-point UNIX
            timestamps.
        :type timestamp_field: str

        :param timeout: Whether or not the Solr connection should
            be persistent or not.
        :type timeout: int

        :param persistent_connection: Maintain a connection between Solr index
            interactions.
        :type persistent_connection: bool

        :param commit_on_set: Immediately commit changes when a vector is set.
        :type commit_on_set: bool

        """
        super(SolrDescriptorElement, self).__init__(type_str, uuid)

        self.type_field = type_field
        self.uuid_field = uuid_field
        self.vector_field = vector_field
        self.timestamp_field = timestamp_field

        self.solr_conn_addr = solr_conn_addr
        self.solr_timeout = timeout
        self.solr_persistent_connection = persistent_connection
        self.solr_commit_on_set = commit_on_set

        self.solr = self._make_solr_inst()

    def __getstate__(self):
        state = super(SolrDescriptorElement, self).__getstate__()
        state.update({
            "type_field": self.type_field,
            "uuid_field": self.uuid_field,
            "vector_field": self.vector_field,
            "timestamp_field": self.timestamp_field,
            "solr_conn_addr": self.solr_conn_addr,
            "solr_persistent_connection": self.solr_persistent_connection,
            "solr_timeout": self.solr_timeout,
            "solr_commit_on_set": self.solr_commit_on_set,
        })
        return state

    def __setstate__(self, state):
        # Support older version of serialization
        if 'type_label' in state:
            self._type_label = state['type_label']
            self._uuid = state['uuid']
        else:
            super(SolrDescriptorElement, self).__setstate__(state)
        self.type_field = state['type_field']
        self.uuid_field = state['uuid_field']
        self.vector_field = state['vector_field']
        self.timestamp_field = state['timestamp_field']
        self.solr_conn_addr = state['solr_conn_addr']
        self.solr_timeout = state['solr_timeout']
        self.solr_persistent_connection = state['solr_persistent_connection']
        self.solr_commit_on_set = state['solr_commit_on_set']

        self.solr = self._make_solr_inst()

    def __repr__(self):
        return super(SolrDescriptorElement, self).__repr__() + \
            '[url: %s, timeout: %d, ' \
            'persistent: %s]' \
            % (self.solr.url, self.solr.timeout, self.solr.persistent)

    def _make_solr_inst(self):
        return solr.Solr(self.solr_conn_addr,
                         persistent=self.solr_persistent_connection,
                         timeout=self.solr_timeout,
                         # debug=True  # This makes things pretty verbose
                         )

    def _base_doc(self):
        t = self.type()
        suuid = str(self.uuid())
        return {
            'id': '-'.join([t, suuid]),
            self.type_field: t,
            self.uuid_field: suuid,
        }

    def _get_existing_doc(self):
        """
        :return: An existing document dict. If there isn't one for our type/uuid
            we return None.
        :rtype: None | dict
        """
        b_doc = self._base_doc()
        r = self.solr.select("id:%s AND %s:%s AND %s:%s"
                             % (b_doc['id'],
                                self.type_field, b_doc[self.type_field],
                                self.uuid_field, b_doc[self.uuid_field]))
        if r.numFound == 1:
            return r.results[0]
        else:
            return None

    def get_config(self):
        return {
            "solr_conn_addr": self.solr_conn_addr,
            "type_field": self.type_field,
            "uuid_field": self.uuid_field,
            "vector_field": self.vector_field,
            "timestamp_field": self.timestamp_field,
            "timeout": self.solr_timeout,
            "persistent_connection": self.solr_persistent_connection,
            "commit_on_set": self.solr_commit_on_set,
        }

    def has_vector(self):
        return bool(self._get_existing_doc())

    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        :param new_vec: New vector to contain.
        :type new_vec: numpy.core.multiarray.ndarray

        :returns: Self.
        :rtype: SolrDescriptorElement

        """
        doc = self._base_doc()
        doc[self.vector_field] = new_vec.tolist()
        doc[self.timestamp_field] = time.time()
        self.solr.add(doc, commit=self.solr_commit_on_set)
        return self

    def vector(self):
        doc = self._get_existing_doc()
        if doc is None:
            return None
        # Vectors stored as lists in solr doc
        return numpy.array(doc[self.vector_field])


DESCRIPTOR_ELEMENT_CLASS = SolrDescriptorElement



from smqtk.tests.utils.test_plugin_get import DummyInterface


class ImplExternal1 (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'external1'+str(val)


class ImplExternal2 (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'external2'+str(val)



import itertools


def check_empty_iterable(iterable, callback, exception_inst):
    """
    Check that the given iterable is not empty, then call the given callback
    function with the reconstructed iterable when it is not empty.

    :param iterable: Iterable to check.
    :type iterable: collections.Iterable

    :param callback: Function to call with the reconstructed, not-empty
        iterable.
    :type callback: (collections.Iterable) -> None

    :param exception_inst: The exception to throw if the iterable is empty
    :type exception_inst: Exception

    """
    i = iter(iterable)
    try:
        first = next(i)
    except StopIteration:
        raise exception_inst
    callback(itertools.chain([first], i))



#!/usr/bin/env python
import argparse
import csv
import itertools
import os
import re

import numpy
import six
from sklearn.model_selection import StratifiedShuffleSplit


class KeyToFilepath(argparse.Action):
    """
    Custom argparse action for parsing out positional class-to-filepath
    arguments.
    """
    re_key2path = re.compile('(\w+)=(.+)', flags=re.UNICODE)

    # noinspection PyUnusedLocal,PyShadowingBuiltins
    def __init__(self, option_strings, dest, nargs=None, const=None,
                 default=None, type=None, choices=None, required=False,
                 help=None, metavar=None):
        """
        Custom constructor to enforce that `nargs` is always `+`.
        """
        super(KeyToFilepath, self).__init__(option_strings,
                                            dest, "+",
                                            const, default, type,
                                            choices, required,
                                            help, metavar)

    # noinspection PyShadowingNames
    def __call__(self, parser, namespace, values, option_string=None):
        d = dict()
        for a in values:
            m = self.re_key2path.match(a)
            if not m:
                raise ValueError("Invalid argument syntax: '%s'" % a)
            cls_name = m.group(1)
            filepath = m.group(2)
            if not os.path.isfile(filepath):
                raise ValueError(
                    "Invalid filepath '%s' given for argument: '%s'"
                    % (filepath, a)
                )
            # Read in UIDs from lines in CSV file
            d[cls_name] = filepath
        setattr(namespace, self.dest, d)


def cli_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('cls_to_cmdProcessedCsv',
                        nargs='+',
                        help="Series of `label=filepath` arguments where we "
                             "interpret the string value before the `=` sign "
                             "as the class label and the value after to be the "
                             "path to the `compute_many_descriptors` output "
                             "CSV of files successfully processed.",
                        action=KeyToFilepath)
    parser.add_argument('-t', '--test-percent',
                        type=float,
                        default=0.3,
                        help="Percentage of images per class to split for "
                             "testing. Should be in the [0,1] range. Selects "
                             "~30%% by default.")
    parser.add_argument('--rand-state', type=int, default=0,
                        help='Random state initialization integer. Default is '
                             '0.')
    parser.add_argument('--output-base',
                        default="classifier",
                        help="String base to output files. We will generate 3 "
                             "files: '<>.all_uuids.csv', '<>.train_uuids.csv' "
                             "and '<>.test_uuids.csv'. "
                             "Default is 'classifier'.")
    return parser


def main():
    args = cli_parser().parse_args()

    TEST_PERCENT = args.test_percent
    RAND_STATE = args.rand_state
    OUTPUT_BASE = args.output_base
    CLS_TO_FILEPATH = args.cls_to_cmdProcessedCsv

    # Parse CSV files associated to classes
    cls_uuids = {}
    for cls, filepath in six.iteritems(CLS_TO_FILEPATH):
        cls_uuids[cls] = sorted({r[1] for r in csv.reader(open(filepath))})

    cls_list = sorted(cls_uuids)
    all_label, all_uuids = \
        zip(*[(cls_name, uuid)
              for cls_name in cls_list
              for uuid in cls_uuids[cls_name]])
    # Transform into numpy array for multi-index access later
    all_label = numpy.array(all_label)
    all_uuids = numpy.array(all_uuids)

    # ``n_splits=1``  -- Only make one train/test split
    sss = StratifiedShuffleSplit(n_splits=1, test_size=TEST_PERCENT,
                                 random_state=RAND_STATE)

    # Get array of index position values of ``all_uuids`` of uuids to use for
    # train and test sets, respectively.
    train_index, test_index = \
        iter(sss.split(numpy.zeros(len(all_label)), all_label)).next()
    uuids_train, uuids_test = all_uuids[train_index], all_uuids[test_index]
    label_train, label_test = all_label[train_index], all_label[test_index]

    print("Train:")
    for cls_label in cls_list:
        cnt = label_train.tolist().count(cls_label)
        print("- %s:\t%d\t(~%.2f %% of total class examples)"
              % (cls_label, cnt, float(cnt) / len(cls_uuids[cls_label]) * 100))
    print("Test:")
    for cls_label in cls_list:
        cnt = label_test.tolist().count(cls_label)
        print("- %s:\t%d\t(~%.2f %% of total class examples)"
              % (cls_label, cnt, float(cnt) / len(cls_uuids[cls_label]) * 100))

    # Save out files for use with ``classifier_model_validation``
    with open('%s.all_uuids.csv' % OUTPUT_BASE, 'w') as f:
        w = csv.writer(f)
        for uuid, label in itertools.izip(all_uuids, all_label):
            w.writerow([uuid, label])

    with open('%s.train_uuids.csv' % OUTPUT_BASE, 'w') as f:
        w = csv.writer(f)
        for uuid, label in itertools.izip(uuids_train, label_train):
            w.writerow([uuid, label])

    with open('%s.test_uuids.csv' % OUTPUT_BASE, 'w') as f:
        w = csv.writer(f)
        for uuid, label in itertools.izip(uuids_test, label_test):
            w.writerow([uuid, label])


if __name__ == '__main__':
    main()



import mimetypes
import multiprocessing
import os

import flask
import requests

from smqtk.algorithms.descriptor_generator import get_descriptor_generator_impls
from smqtk.representation import DescriptorElementFactory
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.data_element.url_element import DataUrlElement
from smqtk.utils import SimpleTimer
from smqtk.utils import plugin
from smqtk.utils import merge_dict
from smqtk.web import SmqtkWebApp


MIMETYPES = mimetypes.MimeTypes()


SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


class DescriptorServiceServer (SmqtkWebApp):
    """
    Simple server that takes in a specification of the following form:

        /<descriptor_type>/<uri>[?...]

    See the docstring for the ``compute_descriptor()`` method for complete rules
    on how to form a calling URL.

    Computes the requested descriptor for the given file and returns that via
    a JSON structure.

    Standard return JSON::

        {
            "success": <bool>,
            "descriptor": [ <float>, ... ]
            "message": <string>,
            "reference_uri": <uri>
        }

    Additional Configuration


    .. note:: We will look for an environment variable
              `DescriptorService_CONFIG` for a string file path to an additional
              JSON configuration file to consider.

    """

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(DescriptorServiceServer, cls).get_default_config()
        merge_dict(c, {
            "descriptor_factory": DescriptorElementFactory.get_default_config(),
            "descriptor_generators": {
                "example": plugin.make_config(get_descriptor_generator_impls())
            }
        })
        return c

    def __init__(self, json_config):
        """
        Initialize application based of supplied JSON configuration

        :param json_config: JSON configuration dictionary
        :type json_config: dict

        """
        super(DescriptorServiceServer, self).__init__(json_config)

        # Descriptor factory setup
        self._log.info("Initializing DescriptorElementFactory")
        self.descr_elem_factory = DescriptorElementFactory.from_config(
            self.json_config['descriptor_factory']
        )

        # Descriptor generator configuration labels
        #: :type: dict[str, dict]
        self.generator_label_configs = self.json_config['descriptor_generators']

        # Cache of DescriptorGenerator instances so we don't have to
        # continuously initialize them as we get requests.
        self.descriptor_cache = {}
        self.descriptor_cache_lock = multiprocessing.RLock()

        @self.route("/")
        def list_ingest_labels():
            return flask.jsonify({
                "labels": sorted(self.generator_label_configs)
            })

        @self.route("/all/content_types")
        def all_content_types():
            """
            Of available descriptors, what content types are processable, and
            what types are associated to which available descriptor generator.
            """
            all_types = set()
            # Mapping of configuration label to content types that generator
            # can handle
            r = {}
            for l in self.generator_label_configs:
                d = self.get_descriptor_inst(l)
                all_types.update(d.valid_content_types())
                r[l] = sorted(d.valid_content_types())

            return flask.jsonify({
                "all": sorted(all_types),
                "by-label": r
            })

        @self.route("/all/compute/<path:uri>")
        def all_compute(uri):
            """
            Compute descriptors over the specified content for all generators
            that function over the data's content type.

            JSON Return format::

                {
                    "success": <bool>

                    "content_type": <str>

                    "message": <str>

                    "descriptors": {  "<label>":  <list[float]>, ... } | None

                    "reference_uri": <str>
                }

            """
            message = "execution nominal"

            data_elem = None
            try:
                data_elem = self.resolve_data_element(uri)
            except ValueError as ex:
                message = "Failed URI resolution: %s" % str(ex)

            descriptors = {}
            finished_loop = False
            if data_elem:
                for l in self.generator_label_configs:
                    if data_elem.content_type() in \
                            self.get_descriptor_inst(l).valid_content_types():
                        d = None
                        try:
                            d = self.generate_descriptor(data_elem, l)
                        except RuntimeError as ex:
                            message = "Descriptor extraction failure: %s" \
                                      % str(ex)
                        except ValueError as ex:
                            message = "Data content type issue: %s" % str(ex)

                        descriptors[l] = d and d.vector().tolist()
                if not descriptors:
                    message = "No descriptors can handle URI content type: %s" \
                              % data_elem.content_type
                else:
                    finished_loop = True

            return flask.jsonify({
                "success": finished_loop,
                "content_type": data_elem.content_type(),
                "message": message,
                "descriptors": descriptors,
                "reference_uri": uri
            })

        @self.route("/<string:descriptor_label>/<path:uri>")
        def compute_descriptor(descriptor_label, uri):
            """

            Data modes for upload/use::

                - local filepath
                - base64
                - http/s URL

            The following sub-sections detail how different URI's can be used.

            Local Filepath
            --------------

            The URI string must be prefixed with ``file://``, followed by the
            full path to the data file to describe.

            Base 64 data
            ------------

            The URI string must be prefixed with "base64://", followed by the
            base64 encoded string. This mode also requires an additional
            ``?content_type=`` to provide data content type information. This
            mode saves the encoded data to temporary file for processing.

            HTTP/S address
            --------------

            This is the default mode when the URI prefix is none of the above.
            This uses the requests module to locally download a data file
            for processing.

            JSON Return format::

                {
                    "success": <bool>

                    "message": <str>

                    "descriptor": <None|list[float]>

                    "reference_uri": <str>
                }

            :type descriptor_label: str
            :type uri: str

            """
            message = "execution nominal"
            descriptor = None

            de = None
            try:
                de = self.resolve_data_element(uri)
            except ValueError as ex:
                message = "URI resolution issue: %s" % str(ex)

            if de:
                try:
                    descriptor = self.generate_descriptor(de, descriptor_label)
                except RuntimeError as ex:
                    message = "Descriptor extraction failure: %s" % str(ex)
                except ValueError as ex:
                    message = "Data content type issue: %s" % str(ex)

            return flask.jsonify({
                "success": descriptor is not None,
                "message": message,
                "descriptor":
                    (descriptor is not None and descriptor.vector().tolist())
                    or None,
                "reference_uri": uri
            })

    def get_config(self):
        return self.json_config

    def get_descriptor_inst(self, label):
        """
        Get the cached content descriptor instance for a configuration label
        :type label: str
        :rtype: smqtk.descriptor_generator.DescriptorGenerator
        """
        with self.descriptor_cache_lock:
            if label not in self.descriptor_cache:
                self._log.debug("Caching descriptor '%s'", label)
                self.descriptor_cache[label] = \
                    plugin.from_plugin_config(
                    self.generator_label_configs[label],
                        get_descriptor_generator_impls()
                    )

            return self.descriptor_cache[label]

    def resolve_data_element(self, uri):
        """
        Given the URI to some data, resolve it down to a DataElement instance.

        :raises ValueError: Issue with the given URI regarding either URI source
            resolution or data resolution.

        :param uri: URI to data
        :type uri: str
        :return: DataElement instance wrapping given URI to data.
        :rtype: smqtk.representation.DataElement

        """
        # Resolve URI into appropriate DataElement instance
        if uri[:7] == "file://":
            self._log.debug("Given local disk filepath")
            filepath = uri[7:]
            if not os.path.isfile(filepath):
                raise ValueError("File URI did not point to an existing file "
                                 "on disk.")
            else:
                de = DataFileElement(filepath)

        elif uri[:9] == "base64://":
            self._log.debug("Given base64 string")
            content_type = flask.request.args.get('content_type', None)
            self._log.debug("Content type: %s", content_type)
            if not content_type:
                raise ValueError("No content-type with given base64 data")
            else:
                b64str = uri[9:]
                de = DataMemoryElement.from_base64(b64str, content_type)

        else:
            self._log.debug("Given URL")
            try:
                de = DataUrlElement(uri)
            except requests.HTTPError as ex:
                raise ValueError("Failed to initialize URL element due to "
                                 "HTTPError: %s" % str(ex))

        return de

    def generate_descriptor(self, de, cd_label):
        """
        Generate a descriptor for the content pointed to by the given URI using
        the specified descriptor generator.

        :raises ValueError: Content type mismatch given the descriptor generator
        :raises RuntimeError: Descriptor extraction failure.

        :type de: smqtk.representation.DataElement
        :type cd_label: str

        :return: Generated descriptor element instance with vector information.
        :rtype: smqtk.representation.DescriptorElement

        """
        with SimpleTimer("Computing descriptor...", self._log.debug):
            cd = self.get_descriptor_inst(cd_label)
            descriptor = cd.compute_descriptor(de, self.descr_elem_factory)

        return descriptor


APPLICATION_CLASS = DescriptorServiceServer



import unittest

from smqtk.utils import (
    merge_dict
)


class TestMergeConfigs (unittest.TestCase):

    def test_disjoint_update(self):
        a = {
            'a': 1,
            'b': 2,
        }
        b = {
            'c': 3
        }
        expected = {
            'a': 1,
            'b': 2,
            'c': 3,
        }
        merge_dict(a, b)
        self.assertEqual(a, expected)

    def test_subset_merge(self):
        a = {
            'a': 1,
            'b': 2,
        }
        b = {
            'a': 3
        }
        expected = {
            'a': 3,
            'b': 2,
        }
        merge_dict(a, b)
        self.assertEqual(a, expected)

    def test_partial_update(self):
        a = {
            'a': 1,
            'b': 2,
        }
        b = {
            'a': 3,
            'c': 4,
        }
        expected = {
            'a': 3,
            'b': 2,
            'c': 4,
        }
        merge_dict(a, b)
        self.assertEqual(a, expected)

    def test_overrides(self):
        a = {
            'a': 1,
            'b': 2,
        }
        b = {
            'b': {
                'c': 3
            },
        }
        expected = {
            'a': 1,
            'b': {
                'c': 3,
            }
        }
        merge_dict(a, b)
        self.assertEqual(a, expected)

    def test_nested(self):
        a = {
            'a': 1,
            'b': {
                'c': 2,
                'd': {
                    'e': 3
                },
            },
            'f': {
                'g': 4,
                'h': {
                    'i': 5
                }
            },
        }
        b = {
            'b': {'c': 6},
            'f': {'h': {'i': 7}},
            'j': 8
        }
        expected = {
            'a': 1,
            'b': {
                'c': 6,
                'd': {
                    'e': 3
                },
            },
            'f': {
                'g': 4,
                'h': {
                    'i': 7
                }
            },
            'j': 8,
        }
        merge_dict(a, b)
        self.assertEqual(a, expected)



from smqtk.exceptions import ReadOnlyError
from smqtk.representation import DataElement


# attempt to import required modules
try:
    import happybase
    import tika
    from tika import detector as tika_detector
except ImportError:
    happybase = None
    tika = None
    tika_detector = None


class HBaseDataElement (DataElement):
    """
    Wrapper for binary data contained on an HBase server somewhere. Uses Tika
    content type detection to determine content type of served data.
    """

    @classmethod
    def is_usable(cls):
        return None not in {happybase, tika_detector}

    def __init__(self, element_key, binary_column, hbase_address,
                 hbase_table, timeout=10000):
        """
        Create a new HBase data element wrapper/reference.

        :param element_key: Key to the table row containing the binary data
        :type element_key: str

        :param binary_column: Name of the column that contains the binary
            data.
        :type binary_column: str

        :param hbase_address: Address of the HBase server. This should at least
            be the hostname of the server. This might also take a ":port"
            suffix?
        :type hbase_address: str

        :param hbase_table: Name of the table the content is contained in.
        :param timeout:
        :return:
        """
        super(HBaseDataElement, self).__init__()

        self.element_key = element_key
        self.binary_column = binary_column
        self.hbase_address = hbase_address
        self.hbase_table = hbase_table
        self.timeout = int(timeout)

        self._binary_ct_cache = None

    def __repr__(self):
        return super(HBaseDataElement, self).__repr__() + \
            "{key: %s, bin_col: %s, hbase_addr: %s, hbase_table: %s, " \
            "timeout: %d}" % (
                self.element_key, self.binary_column, self.hbase_address,
                self.hbase_table, self.timeout
            )

    def get_config(self):
        return {
            "element_key": self.element_key,
            "binary_column": self.binary_column,
            "hbase_address": self.hbase_address,
            "hbase_table": self.hbase_table,
            "timeout": self.timeout,
        }

    def content_type(self):
        if self._binary_ct_cache is None:
            self._binary_ct_cache = tika_detector.from_buffer(self.get_bytes())
        return self._binary_ct_cache

    def _new_hbase_table_connection(self):
        return happybase.Connection(self.hbase_address, timeout=self.timeout)\
            .table(self.hbase_table)

    def is_empty(self):
        """
        Check if this element contains no bytes.

        :return: If this element contains 0 bytes.
        :rtype: bool

        """
        # naive impl for now
        return len(self.get_bytes()) == 0

    def get_bytes(self):
        table = self._new_hbase_table_connection()
        r = table.row(self.element_key, columns=[self.binary_column])
        return r[self.binary_column]

    def writable(self):
        """
        :return: if this instance supports setting bytes.
        :rtype: bool
        """
        # No write support (yet) for HBase elements
        return False

    def set_bytes(self, b):
        """
        Set bytes to this data element in the form of a string.

        Not all implementations may support setting bytes (writing). See the
        ``writable`` method.

        :param b: bytes to set.
        :type b: str

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        raise ReadOnlyError("HBase elements cannot write data.")


DATA_ELEMENT_CLASS = HBaseDataElement



from __future__ import division, print_function
import abc
import os
import unittest

from smqtk.utils.plugin import Pluggable, get_plugins, OS_ENV_PATH_SEP


class DummyInterface (Pluggable):

    @abc.abstractmethod
    def inst_method(self, val):
        """
        dummy abstract function
        """


class TestGetPluginGeneric (unittest.TestCase):

    INTERNAL_PLUGIN_DIR = os.path.join(os.path.dirname(__file__),
                                       "test_plugin_dir",
                                       "internal_plugins")

    INTERNAL_PLUGIN_MOD_PATH = \
        'smqtk.tests.utils.test_plugin_dir.internal_plugins'

    ENV_VAR = "TEST_PLUGIN_MODULE_PATH"
    HELP_VAR = "TEST_PLUGIN_CLASS"

    EXT_MOD_1 = 'smqtk.tests.utils.test_plugin_dir.external_1'
    EXT_MOD_2 = 'smqtk.tests.utils.test_plugin_dir.external_2'

    @classmethod
    def get_dummy_plugins(cls):
        return get_plugins(cls.INTERNAL_PLUGIN_MOD_PATH,
                           cls.INTERNAL_PLUGIN_DIR, cls.ENV_VAR,
                           cls.HELP_VAR, DummyInterface)

    def test_get_internal_modules(self, do_return=False):
        m = self.get_dummy_plugins()
        self.assertIn('ImplFoo', m)
        self.assertIn('ImplBar', m)
        self.assertIn('ImplDoExport', m)

        self.assertEqual(m['ImplFoo']().inst_method('a'), 'fooa')
        self.assertEqual(m['ImplBar']().inst_method('b'), 'barb')
        self.assertEqual(m['ImplDoExport']().inst_method('c'), 'doExportc')

        self.assertNotIn('ImplNotUsable', m)
        self.assertNotIn('SomethingElse', m)
        self.assertNotIn('ImplNoExport', m)
        self.assertNotIn('ImplSkipModule', m)

        if do_return:
            return m

    def test_external_1_only(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = self.EXT_MOD_1
        m = self.test_get_internal_modules(True)

        self.assertIn('ImplExternal1', m)
        self.assertIn('ImplExternal2', m)
        self.assertNotIn('ImplExternal3', m)

        self.assertEqual(m['ImplExternal1']().inst_method('d'), 'external1d')
        self.assertEqual(m['ImplExternal2']().inst_method('e'), 'external2e')

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]

    def test_external_1_with_trailing_sep(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = self.EXT_MOD_1+OS_ENV_PATH_SEP
        m = self.test_get_internal_modules(True)

        self.assertIn('ImplExternal1', m)
        self.assertIn('ImplExternal2', m)
        self.assertNotIn('ImplExternal3', m)

        self.assertEqual(m['ImplExternal1']().inst_method('d'), 'external1d')
        self.assertEqual(m['ImplExternal2']().inst_method('e'), 'external2e')

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]

    def test_external_1_with_leading_sep(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = OS_ENV_PATH_SEP+self.EXT_MOD_1
        m = self.test_get_internal_modules(True)

        self.assertIn('ImplExternal1', m)
        self.assertIn('ImplExternal2', m)
        self.assertNotIn('ImplExternal3', m)

        self.assertEqual(m['ImplExternal1']().inst_method('d'), 'external1d')
        self.assertEqual(m['ImplExternal2']().inst_method('e'), 'external2e')

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]

    def test_external_2_only(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = self.EXT_MOD_2
        m = self.test_get_internal_modules(True)

        self.assertNotIn('ImplExternal1', m)
        self.assertNotIn('ImplExternal2', m)
        self.assertIn('ImplExternal3', m)

        self.assertEqual(m['ImplExternal3']().inst_method('f'), 'external3f')

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]

    def test_external_1_and_2(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = OS_ENV_PATH_SEP.join([self.EXT_MOD_1,
                                                         self.EXT_MOD_2])
        m = self.test_get_internal_modules(True)

        self.assertIn('ImplExternal1', m)
        self.assertIn('ImplExternal2', m)
        self.assertIn('ImplExternal3', m)

        self.assertEqual(m['ImplExternal1']().inst_method('d'), 'external1d')
        self.assertEqual(m['ImplExternal2']().inst_method('e'), 'external2e')
        self.assertEqual(m['ImplExternal3']().inst_method('f'), 'external3f')

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]

    def test_junk_external_mod(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = "This is a junk string"
        # This should skip module it can't find. In this case, the junk string
        # is treated as a python module path, which is invalid. A warning is
        # emitted but the plugin query succeeds as if the invalid chunk didn't
        # exist.
        self.test_get_internal_modules()

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]

    def test_external_1_and_2_and_garbage(self):
        env_orig_value = os.environ.get(self.ENV_VAR, None)

        os.environ[self.ENV_VAR] = OS_ENV_PATH_SEP.join([self.EXT_MOD_1,
                                                         self.EXT_MOD_2,
                                                         'asdgasfhsadf',
                                                         'some thing weird',
                                                         'but still uses sep'])
        m = self.test_get_internal_modules(True)

        self.assertIn('ImplExternal1', m)
        self.assertIn('ImplExternal2', m)
        self.assertIn('ImplExternal3', m)

        self.assertEqual(m['ImplExternal1']().inst_method('d'), 'external1d')
        self.assertEqual(m['ImplExternal2']().inst_method('e'), 'external2e')
        self.assertEqual(m['ImplExternal3']().inst_method('f'), 'external3f')

        if env_orig_value:
            os.environ[self.ENV_VAR] = env_orig_value
        else:
            del os.environ[self.ENV_VAR]



"""
LICENCE
-------
Copyright 2013 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

"""

from six.moves import configparser
from six.moves import StringIO


class SafeConfigCommentParser (configparser.SafeConfigParser, object):
    """
    Adaptation of the configparser.SafeConfigParser object that can also record
    a comment for an option.

    Matched to python 2.6 configparser version at the minimum.

    NOTE: As of right now, comments are not read in from file. They are there
    for creating a configuration object for the intent of writing out to file.
    """

    # Number of characters to limit output printing per line (for comments)
    OUTPUT_LINE_LENGTH = 80

    def __init__(self, defaults=None, dict_type=dict):
        super(SafeConfigCommentParser, self).__init__(defaults, dict_type)

        # We will contain associations of comments to a specific section or
        # option:
        #
        # {
        #   '{section}' : { '__name__' | '{option}' : "comment",
        #                   ... },
        #   ...
        # }
        #
        # If there is no comment associated with an option, there will either
        # be no entry along that path, or that entry will be None.
        #
        self._comments = self._dict()

        # Add slot for DEFAULT section option comments since it won't be added
        # by the user. Also adding an initial null comment to get around bug in
        # comment retrieval.
        self._comments[configparser.DEFAULTSECT] = self._dict()
        self.set_comment(None, configparser.DEFAULTSECT)

    def write(self, fp):
        """
        Override of write method to provide a better formatted output, taking
        into consideration comments for an option.

        Write an .ini-format representation of the configuration state.

        @type fp: file

        """
        def write_comment(comment):
            """ write comment to file
            @type comment: str

            """
            # Split the comment into blocks based on new-line characters.
            # - Adding a new-line at the end of the comment so that, if an user
            #   explicitly placed a '\n' at the end of their comment, it will be
            #   registered as an expected new comment line with nothing on it.
            blocks = (comment+'\n').splitlines()

            # Each new block with start on a new line, directly underneath the
            # previous block.
            for block in blocks:
                segments = block.split(' ')
                line_buff = ''
                block_buff = ''

                while segments:
                    # if the concatenation of the current segment would make the
                    # total line buffer length greater than the output length
                    # limit, dump the line to the block buffer.

                    line_length = len(line_buff) + len(segments[0]) + 1
                    if line_buff and line_length > self.OUTPUT_LINE_LENGTH:
                        # The buffer isn't empty and the next segment won't fit
                        # in the buffer. Dumping out the current buffer into the
                        # block buffer and starting a new line buffer. The '+1'
                        # in the 'if' is to take into account the additional
                        # space to join buffer and the new segment.
                        block_buff += line_buff + '\n'
                        line_buff = ''
                    elif line_buff:
                        # add the segment to the non-empty buffer
                        line_buff = ' '.join([line_buff, segments.pop(0)])
                    else:
                        # Nothing in the buffer. Start a new one with the
                        # starting ;
                        line_buff = ' '.join([';', segments.pop(0)])

                # flush the remaining buffer to the block buffer
                if line_buff:
                    block_buff += line_buff + '\n'

                # Write the block out. Every block should already end in a new-
                # line char because of how lines are added to the buffer.
                # - fp is the file object passed to the parent write method
                fp.write(block_buff)

        if self._defaults:
            fp.write("[%s]\n" % configparser.DEFAULTSECT)
            for (key, value) in self._defaults.items():
                # write comment if there is one for this default key
                has_comment = (
                    self._comments.get(configparser.DEFAULTSECT, False) and
                    self._comments[configparser.DEFAULTSECT].get(key, False)
                )
                if has_comment:
                    write_comment(self._comments[configparser.DEFAULTSECT][key])
                # write key/val
                fp.write("%s = %s\n" % (key, str(value).replace('\n', '\n\t')))
            # spacer before next section heading
            fp.write("\n")

        for section in sorted(self._sections):
            fp.write('\n')

            # write comment for section if there is one
            if self.has_comment(section):
                write_comment(self.get_comment(section))

            # section head
            fp.write("[%s]\n" % section)

            for (key, value) in sorted(self._sections[section].items()):
                if key != "__name__":
                    fp.write("\n")
                    # write comment for pair of there is one
                    if self.has_comment(section, key):
                        write_comment(self.get_comment(section, key))
                    # write key/val
                    fp.write("%s = %s\n" %
                             (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")

    def add_section(self, section, comment=None):
        """
        Create a new section in the configuration with an optional comment.

        Raise DuplicateSectionError if a section by the specified name
        already exists. Raise ValueError if name is DEFAULT or any of it's
        case-insensitive variants.

        @type section:str
        @type comment: str or None

        """
        super(SafeConfigCommentParser, self).add_section(section)

        if comment is not None:
            self._comments[section] = self._dict()
            self._comments[section]['__name__'] = comment

    def set(self, section, option, value=None, comment=None):
        """
        Set an option.  Extend configparser.set: check for string values. Also
        optionally records a comment for the option.

        @type section: str
        @type option: str
        @type value: str or None
        @type comment: str or None

        """
        super(SafeConfigCommentParser, self).set(section, option, value)

        if comment is not None:
            if not section:
                section = configparser.DEFAULTSECT

            if section not in self._comments:
                self._comments[section] = self._dict()

            self._comments[section][self.optionxform(option)] = comment

    def has_comment(self, section, option=None):
        """
        Return if the given section heading or option within a section has an
        associated comment.

        @type section: str
        @type option: str or None
        @rtype: bool

        @raise NoSectionError: The provided section does not exist

        """
        if not section:
            section = configparser.DEFAULTSECT

        if not self.has_section(section) and \
                section != configparser.DEFAULTSECT:
            raise configparser.NoSectionError(section)

        if option:
            option = self.optionxform(option)
            if not self.has_option(section, option):
                raise configparser.NoOptionError(option, section)
        else:
            # looking for section comment
            option = '__name__'

        return bool(self._comments.get(section, False)
                    and self._comments[section].get(option, False))

    def get_comment(self, section, option=None):
        """
        Get the comment for a section[.option]

        @type section: str
        @param section: Section heading to check for a comment, or the section
        heading that the target option is located under.

        @type option: str or None
        @param option: If checking for an option's comment, this is the option
        under the given section to check. If checking for a section's comment,
        this should be None.

        @rtype: str or None
        @return: The section or option comment if there is one, or None if there
        is no comment for the specified section or option.

        """
        if not section:
            section = configparser.DEFAULTSECT

        if not self.has_section(section) and \
                section != configparser.DEFAULTSECT:
            raise configparser.NoSectionError(section)

        if option:
            option = self.optionxform(option)
            if not self.has_option(section, option):
                raise configparser.NoOptionError(option, section)

        else:
            # looking for section comment
            option = '__name__'

        # Combined statement to handle both section and option requests.
        return (self._comments.get(section, None) and
                self._comments[section].get(option, None))

    def set_comment(self, comment, section, option=None):
        """
        Set the comment for a section or option

        @type comment: str or None
        @type section: str
        @type option: str or None

        """
        if not section:
            section = configparser.DEFAULTSECT

        if not self.has_section(section) and \
                section != configparser.DEFAULTSECT:
            raise configparser.NoSectionError(section)

        if section not in self._comments:
            self._comments[section] = self._dict()

        if option:
            option = self.optionxform(option)
            if not self.has_option(section, option):
                raise configparser.NoOptionError(option, section)
        else:
            # setting section comment
            option = '__name__'

        self._comments[section][option] = comment

    def remove_comment(self, section, option=None):
        """
        Remove the comment from a section or option.

        @type section: str
        @param section: The section to remove from.

        @type option: str or None
        @param option: The option to remove from or None to remove a section's
        comment.

        @rtype: bool
        @return: True if a comment was removed and False if no comment was
        removed.

        """
        if not section:
            section = configparser.DEFAULTSECT
        elif not self.has_section(section) and \
                section != configparser.DEFAULTSECT:
            raise configparser.NoSectionError(section)

        if option:
            option = self.optionxform(option)
            if not self.has_option(section, option):
                raise configparser.NoOptionError(option, section)
        else:
            # setting section comment
            option = '__name__'

        del self._comments[section][option]

    def remove_option(self, section, option):
        """
        Remove an option and any comments associated with it.

        @type section: str
        @param section: The section the option is located in.

        @type option: str
        @param option: The option to remove

        @rtype: bool
        @return: True of the option existed under the given section and was
        removed. False if the option didn't exist and couldn't be removed.

        """
        if super(SafeConfigCommentParser, self).remove_option(section, option):
            if option in self._comments[section]:
                del self._comments[section][option]
            return True
        return False

    def remove_section(self, section):
        """
        Remove a file section and any comments associated with it.

        @type section: str
        @param section: The section to remove.

        @rtype: bool
        @return: True of the section existed and was removed. False if the
        section didn't exist and couldn't be removed.

        """
        if super(SafeConfigCommentParser, self).remove_section(section):
            # With the removal of an entire section entry in the comments
            # dictionary
            # self.remove_comment(section)
            if section in self._comments:
                del self._comments[section]
            return True
        return False

    def as_str(self):
        """
        :return: this configuration as a string as would be written to a file.
        :rtype: str

        """
        buff = StringIO()
        # noinspection PyTypeChecker
        # reason: StringIO implements file-like interface.
        self.write(buff)
        return buff.getvalue()

    __str__ = as_str

    def update(self, *configs):
        """
        Update this configuration object with the values of the given, other
        configuration objects. This updates this object in place. Also updates
        comments.

        The order in which configuration objects are given matters, where the
        first provided will be the first that is updated into this object, and
        the last provided will be the last be the last updated into this object.
        If multiple config objects contain the same section option pairs, the
        value in the last configuration object provided will take priority.

        REMEMBER
        --------
        Section options that are set to an empty string is a valid value and
        will cause an overwrite in this config and/or preceding configs in the
        given order.

        :param configs: One or more SafeConfigCommentParser objects
        :type configs: Iterable of SafeConfigCommentParser
        :return: This config object (in-place update)
        :rtype: SafeConfigCommentParser

        """
        for config in configs:
            self.defaults().update(config.defaults())

            for section in config.sections():
                if not self.has_section(section):
                    self.add_section(section, config.get_comment(section))
                elif config.has_comment(section):
                    self.set_comment(config.get_comment(section), section)

                for option in config.options(section):
                    self.set(section, option, config.get(section, option))
                    # Only update the commend if the current config actually has
                    # one.
                    if config.has_comment(section, option):
                        self.set_comment(config.get_comment(section, option),
                                         section, option)



"""
Utility script to transform a set of descriptors, specified by UUID, with
matching class labels, to a test file usable by libSVM utilities for
train/test experiments.

The input CSV file is assumed to be of the format:

    uuid,label
    ...

This is the same as the format requested for other scripts like
``classifier_model_validation.py``.

This is very useful for searching for -c and -g parameter values for a
training sample of data using the ``tools/grid.py`` script, found in the
libSVM source tree. For example:

    <smqtk_source>/TPL/libsvm-3.1-custom/tools/grid.py \
        -log2c -5,15,2 -log2c 3,-15,-2 -v 5 -out libsvm.grid.out \
        -png libsvm.grid.png -t 0 -w1 3.46713615023 -w2 12.2613240418 \
        output_of_this_script.txt
"""

import csv
import logging
import six
from six.moves import zip

import six

from smqtk.representation import (
    get_descriptor_index_impls,
)
from smqtk.utils import (
    bin_utils,
    plugin,
)


def default_config():
    return {
        'plugins': {
            'descriptor_index':
                plugin.make_config(get_descriptor_index_impls()),
        }
    }


def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)

    g_io = parser.add_argument_group("IO Options")
    g_io.add_argument('-f', metavar='PATH',
                      help='Path to the csv file mapping descriptor UUIDs to '
                           'their class label. String labels are transformed '
                           'into integers for libSVM. Integers start at 1 '
                           'and are applied in the order that labels are '
                           'seen in this input file.')
    g_io.add_argument('-o', metavar='PATH',
                      help='Path to the output file to write libSVM labeled '
                           'descriptors to.')
    return parser


def main():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(
        config['plugins']['descriptor_index'],
        get_descriptor_index_impls()
    )

    labels_filepath = args.f
    output_filepath = args.o

    # Run through labeled UUIDs in input file, getting the descriptor from the
    # configured index, applying the appropriate integer label and then writing
    # the formatted line out to the output file.
    input_uuid_labels = csv.reader(open(labels_filepath))

    with open(output_filepath, 'w') as ofile:
        label2int = {}
        next_int = 1
        uuids, labels = list(zip(*input_uuid_labels))

        log.info("Scanning input descriptors and labels")
        for i, (l, d) in enumerate(
                    zip(labels, descriptor_index.get_many_descriptors(uuids))):
            log.debug("%d %s", i, d.uuid())
            if l not in label2int:
                label2int[l] = next_int
                next_int += 1
            ofile.write(
                "%d " % label2int[l] +
                ' '.join(["%d:%.12f" % (j+1, f)
                          for j, f in enumerate(d.vector())
                          if f != 0.0]) +
                '\n'
            )

    log.info("Integer label association:")
    for i, l in sorted((i, l) for l, i in six.iteritems(label2int)):
        log.info('\t%d :: %s', i, l)


if __name__ == '__main__':
    main()



import unittest

from smqtk.algorithms.classifier import SupervisedClassifier


class DummySupervisedClassifier (SupervisedClassifier):

    EXPECTED_LABELS = ['constant']
    EXPECTED_HAS_MODEL = False

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        return {}

    def get_labels(self):
        return self.EXPECTED_LABELS

    def _classify(self, d):
        return {d.uuid(): d.vector().tolist()}

    def has_model(self):
        return self.EXPECTED_HAS_MODEL

    def _train(self, class_examples, **extra_params):
        # Return super-method result in its attempt to unify mappings
        return class_examples


class TestSupervisedClassifierAbstractClass (unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_classifier = DummySupervisedClassifier()

    def test_train_hasModel(self):
        # Calling the train method should fail the class also reports that it
        # already has a model. Shouldn't matter what is passed to the method
        # (or lack of things passed to the method).
        self.test_classifier.EXPECTED_HAS_MODEL = True
        self.assertRaises(
            RuntimeError,
            self.test_classifier.train, {}
        )

    #
    # Testing train abstract function functionality. Method currently does not
    # care what the value for labels are.
    #

    def test_train_noModel_noExamples(self):
        self.test_classifier.EXPECTED_HAS_MODEL = False
        self.assertRaises(
            ValueError,
            self.test_classifier.train, {}
        )

    def test_train_noModel_oneExample_classExamples(self):
        self.test_classifier.EXPECTED_HAS_MODEL = False
        input_class_examples = {
            'label_1': [0, 1, 2],
        }
        self.assertRaises(
            ValueError,
            self.test_classifier.train, input_class_examples
        )

    def test_train_noModel_classExamples_only(self):
        self.test_classifier.EXPECTED_HAS_MODEL = False
        input_class_examples = {
            'label_1': [0, 1, 2, 3],
            'label_2': [3, 4, 5, 6],
            'label_3': [8, 9, 10, 11],
            'special symbolLabel +here': [5, 1, 76, 8, 9, 2, 5],
        }
        # Intentionally not passing DescriptorElements here.
        # noinspection PyTypeChecker
        m = self.test_classifier.train(class_examples=input_class_examples)
        self.assertEqual(m, input_class_examples)



import mimetypes as pymimetypes
import threading


MT_CACHE_LOCK = threading.RLock()
MT_CACHE = None


def get_mimetypes():
    """
    Get the singleton SMQTK-adjusted mimetypes instance.

    This is desired due to the presence of some odd mimetype and extension
    conversion, e.g. image/jpeg -> .jfif.

    :return: SMQTK singleton MimeTypes instance.
    :rtype: mimetypes.MimeTypes

    """
    global MT_CACHE
    with MT_CACHE_LOCK:
        if MT_CACHE is None:
            MT_CACHE = pymimetypes.MimeTypes()
            # Remove weird extensions from map for getting extensions-by-type
            for jpg_ext in ['.jfif', '.jpe']:
                if jpg_ext in MT_CACHE.types_map_inv[1]['image/jpeg']:
                    MT_CACHE.types_map_inv[1]['image/jpeg'].remove(jpg_ext)
    return MT_CACHE



"""
LICENCE
-------
Copyright 2013-2015 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.
"""

import logging

import numpy as np
from six.moves import range

from smqtk.utils import ReadWriteLock
from smqtk.utils import SimpleTimer
from smqtk.utils.parallel import parallel_map
from smqtk.utils.bin_utils import report_progress


def compute_distance_kernel(m, dist_func, row_wise=False, parallel=True):
    """
    Method for computing the distance kernel of an array of vectors given a
    distance function that works on two supplied 1D arrays.

    For a valid distance function interface, see
    ``smqtk.utils.distance_functions.histogram_intersection_distance2``.

    :param m: An array of vectors to compute the pairwise distance kernel for.
    :type m: numpy.ndarray

    :param dist_func: Distance function
    :type dist_func: (ndarray, ndarray) -> ndarray[float] | float

    :param row_wise: If the given distance function can take a vector and a
        matrix, and computes pair-wise distances, returning a vector of
        distances between the given vector and each row of the matrix.
    :type row_wise: bool

    :param parallel: If distances should be calculated in parallel. This is true
        by default.
    :type parallel: bool

    :return: Computed symmetric distance kernel
    :rtype: numpy.ndarray

    """
    log = logging.getLogger(__name__)

    if m.ndim == 1:
        m = m[np.newaxis]

    log.info("Computing distance kernel")
    side = m.shape[0]
    mat = np.ndarray((side, side), dtype=float)

    s = [0] * 7
    if row_wise:
        log.debug("Computing row-wise distances")
        # For all rows except the last one. We'll have computed all distances by
        # the time reach m[side-1]
        if parallel:
            # noinspection PyShadowingNames
            def work_func(i):
                mat[i, i] = dist_func(m[i], m[i])
                if i < (side - 1):
                    mat[i + 1:, i] = mat[i, i + 1:] = dist_func(m[i, :],
                                                                m[i + 1:, :])
            # Using threading for in-place modification
            s = [0] * 7
            for _ in parallel_map(work_func, range(side),
                                  use_multiprocessing=False):
                report_progress(log.debug, s, 1.)
        else:
            for i in range(side):
                # Compute col/row wise distances
                mat[i, i] = dist_func(m[i], m[i])
                if i < (side-1):
                    mat[i+1:, i] = mat[i, i+1:] = dist_func(m[i, :], m[i+1:, :])
                report_progress(log.debug, s, 1.)
    else:
        log.debug("Computing element-wise distances")
        if parallel:
            # noinspection PyShadowingNames
            def work_func(i):
                mat[i, i] = dist_func(m[i], m[i])
                # cols to the left of diagonal index for this row
                for j in range(i):
                    mat[i, j] = mat[j, i] = dist_func(m[i], m[j])
            # Using threading for in-place modification
            for _ in parallel_map(work_func, range(side),
                                  use_multiprocessing=False):
                report_progress(log.debug, s, 1.)
        else:
            for i in range(side):
                mat[i, i] = dist_func(m[i], m[i])
                # cols to the left of diagonal index for this row
                for j in range(i):
                    mat[i, j] = mat[j, i] = dist_func(m[i], m[j])
                report_progress(log.debug, s, 1.)

    return mat


def compute_distance_matrix(m1, m2, dist_func, row_wise=False):
    """
    Function for computing the pair-wise distance matrix between two arrays of
    vectors. Both matrices must have the same number of columns.
    """
    if m1.ndim == 1:
        m1 = m1[np.newaxis]
    if m2.ndim == 1:
        m2 = m2[np.newaxis]
    k = np.ndarray((m1.shape[0], m2.shape[0]), dtype=float)
    if row_wise:
        # row wise
        for i in range(m1.shape[0]):
            k[i, :] = dist_func(m1[i], m2)
    else:
        for i in range(m1.shape[0]):
            for j in range(m2.shape[0]):
                k[i, j] = dist_func(m1[i], m2[j])
    return k


class DistanceKernel (object):
    """
    Feature Distance Kernel object.

    This class allows the kernel to either be symmetric or not. If it is
    symmetric, the ``symmetric_submatrix`` function becomes available.

    Intended to be used with ProxyManager proxy objects (given at
    construction)

    MONKEY PATCHING:
    When using this object directly (not using the ProxyManager stuff) and
    sending it over pipes, the ReadWriteLock needs to be monkey patched out (the
    multiprocessing.Condition variable doesn't play nicely). Need to set an
    instance of a DummyRWLock to the DistanceKernel._rw_lock property. For
    example:

        ...
        dk = ...
        dk._rw_lock = DummyRWLock()
        <send dk into a pipe>
        ...

    """

    @classmethod
    def construct_symmetric_from_files(cls, id_vector_file, kernel_mat_file,
                                       bg_flags_file=None):
        """
        Construct a symmetric DistanceKernel object, requiring a background
        flags file to denote clip IDs that are to be treated as background
        clips (required to activate symmetric_submatrix function). Such a

        DistanceKernel is usually used with event learning and should be
        provided a background flags file also.

        :param id_vector_file: File containing the numpy.savetxt(...) output of
            clip ID values in the order in which they associate to the rows of
            the kernel matrix.
        :type id_vector_file: str
        :param kernel_mat_file: File containing the kernel matrix as saved by
            numpy.save(...) (saved as an ndarray, converted to matrix on load).
        :type kernel_mat_file: str
        :param bg_flags_file: Optional file containing output of
            numpy.savetxt(...) where each index maps a row index of the kernel
            to whether or not the associated clip ID should be considered a
            background video or not.
        :type bg_flags_file: str
        :return: Symmetric DistanceKernel constructed with the data provided in
            the provided files.
        :rtype: DistanceKernel

        """
        clip_ids = np.array(np.loadtxt(id_vector_file))
        # noinspection PyCallingNonCallable
        kernel_mat = np.matrix(np.load(kernel_mat_file))

        if bg_flags_file is not None:
            bg_flags = np.array(np.loadtxt(bg_flags_file))
            bg_clips = np.array([clip_ids[i]
                                 for i, e in enumerate(bg_flags)
                                 if e])
        else:
            bg_clips = None

        return DistanceKernel(clip_ids, clip_ids, kernel_mat, set(bg_clips))

    @classmethod
    def construct_asymmetric_from_files(cls, row_ids_file, col_ids_file,
                                        kernel_mat_file):
        """
        Construct an asymmetric DistanceKernel object, usually used for archive
        searches.

        No option for providing background clip IDs as asymmetric kernels are
        NOT used for learning purposes.

        :param row_ids_file: File containing the numpy.savetxt(...) output of
            clip ID values in the order in which they associate to the rows of
            the given kernel matrix.
        :type row_ids_file: str
        :param col_ids_file: File containing the numpy.savetxt(...) output of
            clip ID values in the order in which they associate to the columns
            of the given kernel matrix.
        :type col_ids_file: str
        :param kernel_mat_file: File containing the kernel matrix as saved by
            numpy.save(...) (saved as an ndarray, converted to matrix on load).
        :type kernel_mat_file: str
        :return: Asymmetric DistanceKernel constructed with the data provided in
            the provided files.
        :rtype: DistanceKernel

        """
        row_cids = np.array(np.loadtxt(row_ids_file))
        col_cids = np.array(np.loadtxt(col_ids_file))
        # noinspection PyCallingNonCallable
        kernel_mat = np.matrix(np.load(kernel_mat_file))
        return DistanceKernel(row_cids, col_cids, kernel_mat)

    @property
    def _log(self):
        return logging.getLogger('.'.join([self.__module__,
                                           self.__class__.__name__]))

    def __init__(self, row_id_index_map, col_id_index_map, kernel_mat,
                 bg_clip_ids=None, rw_lock=None):
        """
        Initialize the kernel matrix. The initialization values will more than
        likely be proxies to np.matrix objects.

        The ``bg_clip_ids`` array may be given when this kernel matrix is to be
        a square, symmetric kernel and activates the use of the
        ``symmetric_submatrix`` method. This array must list clip IDs that are
        to be considered "background" IDs, or clips that are to always be
        considered negative. These clip IDs must be included in symmetric
        sub-matrices.

        This array must be the same dimension as
        the row and column indices, containing boolean flags. These flags mark
        that the clip ID found at the same index in the edge ID maps should be
        considered a "background" ID, or one that is always treated as a
        negative. This is for the stipulation in the symmetric_submatrix method
        that it always includes the background ID set in the submatrix.

        :param row_id_index_map: Array of clip IDs associated to row indices.
            Contents will be treated as ints.
        :type row_id_index_map: ndarray of int
        :param col_id_index_map: Array of clip IDs associated to row indices.
            Contents will be treated as ints.
        :type col_id_index_map: ndarray of int
        :param kernel_mat: Kernel data matrix.
        :type kernel_mat: matrix
        :param bg_clip_ids: Optional array of boolean flags, marking whether an
            index should be considered a "background" video. Contents will be
            treated as ints.
        :type bg_clip_ids: None | set[int]
        :param rw_lock: Read-Write lock for data provided. This should be
            provided if the any of the data is shared with other objects/
            sources. If this is given None (default), then a lock is created.
        :type rw_lock: ReadWriteLock or None

        """
        # TODO: Possibly add checks for the id arrays like there is for the
        #       bgclipid array (int-able contents)
        assert row_id_index_map.shape[0] == kernel_mat.shape[0], \
            "Length of row index map and kernel row count did not match! " \
            "(row index map: %d, kernel row count: %d)" \
            % (row_id_index_map.shape[0], kernel_mat.shape[0])
        assert col_id_index_map.shape[0] == kernel_mat.shape[1], \
            "Length of col index map and kernel col count did not match! " \
            "(col index map: %d, kernel col count: %d)" \
            % (col_id_index_map.shape[0], kernel_mat.shape[1])

        self._row_id_index_map = row_id_index_map
        self._col_id_index_map = col_id_index_map
        self._kernel = kernel_mat

        assert ((bg_clip_ids is None)
                or isinstance(bg_clip_ids, (set, frozenset))), \
            "Must either give None or a set for the bg_clip_ids " \
            "vector. Got: %s" % type(bg_clip_ids)
        self._bg_cid_set = bg_clip_ids
        if bg_clip_ids is not None:
            try:
                [int(e) for e in bg_clip_ids]
            except Exception:
                raise ValueError("Not all of the contents of of bg_clip_ids "
                                 "could be treated as ints!")

        if rw_lock:
            assert isinstance(rw_lock, ReadWriteLock), \
                "Did not receive valid istance of RW Lock. Got '%s'" \
                % type(rw_lock)
            self._rw_lock = rw_lock
        else:
            self._rw_lock = ReadWriteLock()

    def get_lock(self):
        """
        :return: This object's read/write lock.
        :rtype: ReadWriteLock
        """
        return self._rw_lock

    def row_id_map(self):
        """
        :return: Row index-to-clipID map
        :rtype: ndarray
        """
        with self.get_lock().read_lock():
            return self._row_id_index_map

    def col_id_map(self):
        """
        :return: Column index-to-clipID map
        :rtype: ndarray
        """
        with self.get_lock().read_lock():
            return self._col_id_index_map

    def get_kernel_matrix(self):
        """
        RETURNED OBJECTS NOT THREAD/PROCESS SAFE. Once retrieved, if
        matrix may be modified by another thread/process

        :return: The underlying kernel matrix.
        :rtype: matrix

        """
        with self.get_lock().read_lock():
            return self._kernel

    def get_background_ids(self):
        """
        RETURNED OBJECTS NOT THREAD/PROCESS SAFE

        :return: The set of background clip IDs. May be None if there was no
            background set initialized.
        :rtype: None or frozenset

        """
        with self.get_lock().read_lock():
            return frozenset(self._bg_cid_set) \
                if self._bg_cid_set is not None \
                else frozenset()

    def is_symmetric(self):
        """
        :return: True if this is a square kernel matrix. This means that clip
            IDs along the row and column axes are the same and in the same order
            (starting from [0,0] and moving outwards).
        :rtype: bool

        """
        with self._rw_lock.read_lock():
            # Doing shape equality short circuit because the return value of
            # numpy.array equality changes depending on this condition, meaning
            # the use of the ...all() member function on the result is not
            # universally possible (i.e. when it returns a bool value when
            # shapes are not equal).

            # noinspection PyUnresolvedReferences
            return (self._row_id_index_map.shape == self._col_id_index_map.shape
                    and
                    (self._row_id_index_map == self._col_id_index_map).all())

    def symmetric_submatrix(self, *clip_ids):
        """
        Return a symmetric sub NxN matrix of the total distance kernel based on
        the clip IDs provided. The background clips will always be included in
        the matrix if this DistanceKernel was constructed with a list of
        background clip IDs.

        Clip IDs provided will be assumed non-background, or positive
        event examples. If the clip ID of a background video is provided as an
        argument, we will reconsider it as a non-background video in the
        returned index-to-is-background mapping (tuple).

        Note: The matrix returned will always be a new instance and not set up
        to use shared memory. When directly used with shared memory objects, it
        will be passed by value, not by reference.

        :param clip_ids: Integer clip IDs to include in the returned matrix. The
            returned matrix will contain all background clip IDs.
        :type clip_ids: Iterable of int
        :return: The index-to-clipID map (tuple), the index-to-is-background map
            (tuple) and the symmetric NxN submatrix, where N is the number of
            clip IDs provided as arguments plus the number of background IDs,
            minus the overlap between those two sets.
        :rtype: tuple of int, tuple of bool, numpy.matrixlib.defmatrix.matrix

        """
        with self._rw_lock.read_lock():
            with SimpleTimer("Checking inputs", self._log.debug):
                if not self.is_symmetric():
                    raise RuntimeError("Cannot get a symmetric sub-matrix if "
                                       "the kernel is not square!")
                # DEPRECATED: Allowing the use of this method without explicitly
                #             providing background cIDs. This object will
                #             probably not ever be used this way, but there's no
                #             reason to explicitly disallow it.
                # if self._bg_cid_vec is None:
                #     raise RuntimeError("Cannot create the square submatrix "
                #                        "without the background flag vector!")

                try:
                    clip_ids = [int(e) for e in clip_ids]
                except ValueError:
                    # Re-raise with more informative exception.
                    raise ValueError("Not all clip IDs could be used as ints!")

                id_diff = set(clip_ids).difference(self._row_id_index_map)
                assert not id_diff, \
                    "Not all clip IDs provided are represented in this " \
                    "distance kernel matrix! (difference: %s)" \
                    % id_diff
                del id_diff

            with SimpleTimer("Computing union of BG clips and provided IDs",
                             self._log.debug):
                if self._bg_cid_set is not None:
                    all_cids = self._bg_cid_set.union(clip_ids)
                else:
                    all_cids = set(clip_ids)

            # Reorder the given clip IDs so that they are in the same relative
            # order as the kernel matrix edges.
            focus_indices = []
            focus_clipids = []
            for idx, cid in enumerate(self._row_id_index_map):
                if (cid in all_cids) and (cid not in focus_clipids):
                    focus_indices.append(idx)
                    focus_clipids.append(cid)

            # index-to-isBG map for return
            # -> IDs provided as arguments are to be considered non-background,
            # even if a the ID is in the background set. All other IDs in the
            # union then must be from the background set.
            focus_id2isbg = []
            for idx in focus_indices:
                cid = self._row_id_index_map[idx]
                focus_id2isbg.append(False if cid in clip_ids else True)

            ret_mat = self._kernel[focus_indices, :][:, focus_indices]
            return focus_clipids, focus_id2isbg, ret_mat

    # noinspection PyPep8Naming
    def extract_rows(self, *clipID_or_IDs):
        """
        Find and return the v-stacked distance vectors, in kernel row order
        (i.e. not in the order given as arguments), of the kernel rows matching
        the given clip IDs.

        Note: The matrix returned will always be a new instance and not set up
        to use shared memory. When directly used with shared memory objects, it
        will be passed by value, not by reference.

        :param clipID_or_IDs: The integer clip ID or IDs of which to get the
            distance vectors for.
        :type clipID_or_IDs: int or Iterable of int

        :return: The row-wise index-to-clipID map (tuple), the column-wise
            index-to-clipID map (tuple), and the KxL shape matrix, where K is
            the number of clip IDs given to the method, and L is the width
            (columns) of the distance kernel.
        :rtype: tuple of int, tuple of int, matrix

        """
        with self._rw_lock.read_lock():
            with SimpleTimer("Checking inputs", self._log.debug):
                try:
                    clipID_or_IDs = frozenset(int(e) for e in clipID_or_IDs)
                except Exception as ex:
                    raise ValueError("Not all clip IDs could be used as ints: "
                                     "%s" % str(ex))

                id_diff = clipID_or_IDs.difference(self._row_id_index_map)
                assert not id_diff, \
                    "Not all clip IDs provided are represented in this " \
                    "distance kernel matrix! (difference: %s)" \
                    % id_diff
                del id_diff

            # Reorder the given clip IDs so that they are in the same relative
            # order as the kernel matrix edge order
            with SimpleTimer("Creating focus index/cid sequence",
                             self._log.debug):
                focus_row_indices = []
                focus_row_clipids = []
                for idx, cid in enumerate(self._row_id_index_map):
                    # if ((cid in clipID_or_IDs)
                    #         and (cid not in focus_row_clipids)):
                    if cid in clipID_or_IDs:
                        focus_row_indices.append(idx)
                        focus_row_clipids.append(cid)

            with SimpleTimer("Cropping kernel to focus range", self._log.debug):
                return (
                    tuple(focus_row_clipids),
                    tuple(self._col_id_index_map),
                    self._kernel[focus_row_indices, :]
                )



"""
Compute a descriptor vector for a given data file, outputting the generated
feature vector to standard out, or to an output file if one was specified (in
numpy format).
"""
from __future__ import print_function
import logging
import os

import numpy

from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.representation import DescriptorElementFactory
from smqtk.algorithms.descriptor_generator import get_descriptor_generator_impls
from smqtk.utils import bin_utils, plugin


def default_config():
    return {
        "descriptor_factory": DescriptorElementFactory.get_default_config(),
        "content_descriptor":
            plugin.make_config(get_descriptor_generator_impls()),
    }


def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)

    parser.add_argument('--overwrite',
                        action='store_true', default=False,
                        help="Force descriptor computation even if an "
                             "existing descriptor vector was "
                             "discovered based on the given content "
                             "descriptor type and data combination.")
    parser.add_argument('-o', '--output-filepath',
                        help='Optional path to a file to output '
                             'feature vector to. Otherwise the feature '
                             'vector is printed to standard out. '
                             'Output is saved in numpy binary format '
                             '(.npy suffix recommended).')

    parser.add_argument("input_file",
                        nargs="?",
                        help="Data file to compute descriptor on")

    return parser


def main():
    parser = cli_parser()
    args = parser.parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    output_filepath = args.output_filepath
    overwrite = args.overwrite

    if not args.input_file:
        log.error("Failed to provide an input file path")
        exit(1)
    elif not os.path.isfile(args.input_file):
        log.error("Given path does not point to a file.")
        exit(1)

    input_filepath = args.input_file
    data_element = DataFileElement(input_filepath)

    factory = DescriptorElementFactory.from_config(config['descriptor_factory'])
    #: :type: smqtk.algorithms.descriptor_generator.DescriptorGenerator
    cd = plugin.from_plugin_config(config['content_descriptor'],
                                   get_descriptor_generator_impls())
    descr_elem = cd.compute_descriptor(data_element, factory, overwrite)
    vec = descr_elem.vector()

    if vec is None:
        log.error("Failed to generate a descriptor vector for the input data!")

    if output_filepath:
        numpy.save(output_filepath, vec)
    else:
        # Construct string, because numpy
        s = []
        # noinspection PyTypeChecker
        for f in vec:
            s.append('%15f' % f)
        print(' '.join(s))


if __name__ == "__main__":
    main()



"""
Compute LSH hash codes based on the provided functor on all or specific
descriptors from the configured index given a file-list of UUIDs.

When using an input file-list of UUIDs, we require that the UUIDs of
indexed descriptors be strings, or equality comparable to the UUIDs' string
representation.

We update a key-value store with the results of descriptor hash computation. We
assume the keys of the store are the integer hash values and the values of the
store are ``frozenset`` instances of descriptor UUIDs (hashable-type objects).
We also assume that no other source is concurrently modifying this key-value
store due to the need to modify the values of keys.
"""
import logging
import os

from smqtk.algorithms import (
    get_lsh_functor_impls,
)
from smqtk.compute_functions import compute_hash_codes
from smqtk.representation import (
    get_descriptor_index_impls,
    get_key_value_store_impls,
)
from smqtk.utils import (
    bin_utils,
    plugin,
)


def uuids_for_processing(uuids, hash2uuids):
    """
    Determine descriptor UUIDs that need processing based on what's already in
    the given ``hash2uuids`` mapping, returning UUIDs that need processing.

    :param uuids: Iterable of descriptor UUIDs.
    :type uuids:

    :param hash2uuids: Existing mapping of computed hash codes to the UUIDs
        of descriptors that generated the hash.
    :type hash2uuids: smqtk.representation.KeyValueStore

    :return: Iterator over UUIDs to process
    :rtype: __generator[collections.Hashable]

    """
    log = logging.getLogger(__name__)
    already_there = frozenset(v for vs in hash2uuids.values() for v in vs)
    skipped = 0
    for uuid in uuids:
        if uuid not in already_there:
            yield uuid
        else:
            skipped += 1
    log.debug("Skipped %d UUIDs already represented in previous hash table",
              skipped)


def default_config():
    return {
        "utility": {
            "report_interval": 1.0,
            "use_multiprocessing": False,
        },
        "plugins": {
            "descriptor_index":
                plugin.make_config(get_descriptor_index_impls()),
            "lsh_functor": plugin.make_config(get_lsh_functor_impls()),
            "hash2uuid_kvstore":
                plugin.make_config(get_key_value_store_impls()),
        },
    }


def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)

    g_io = parser.add_argument_group("I/O")
    g_io.add_argument("--uuids-list",
                      default=None, metavar="PATH",
                      help='Optional path to a file listing UUIDs of '
                           'descriptors to computed hash codes for. If '
                           'not provided we compute hash codes for all '
                           'descriptors in the configured descriptor index.')

    return parser


def main():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    #
    # Load configuration contents
    #
    uuid_list_filepath = args.uuids_list
    report_interval = config['utility']['report_interval']
    use_multiprocessing = config['utility']['use_multiprocessing']

    #
    # Checking input parameters
    #
    if (uuid_list_filepath is not None) and \
            not os.path.isfile(uuid_list_filepath):
        raise ValueError("UUIDs list file does not exist!")

    #
    # Loading stuff
    #
    log.info("Loading descriptor index")
    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(
        config['plugins']['descriptor_index'],
        get_descriptor_index_impls()
    )
    log.info("Loading LSH functor")
    #: :type: smqtk.algorithms.LshFunctor
    lsh_functor = plugin.from_plugin_config(
        config['plugins']['lsh_functor'],
        get_lsh_functor_impls()
    )
    log.info("Loading Key/Value store")
    #: :type: smqtk.representation.KeyValueStore
    hash2uuids_kvstore = plugin.from_plugin_config(
        config['plugins']['hash2uuid_kvstore'],
        get_key_value_store_impls()
    )

    # Iterate either over what's in the file given, or everything in the
    # configured index.
    def iter_uuids():
        if uuid_list_filepath:
            log.info("Using UUIDs list file")
            with open(uuid_list_filepath) as f:
                for l in f:
                    yield l.strip()
        else:
            log.info("Using all UUIDs resent in descriptor index")
            for k in descriptor_index.keys():
                yield k

    #
    # Compute codes
    #
    log.info("Starting hash code computation")
    kv_update = {}
    for uuid, hash_int in \
            compute_hash_codes(uuids_for_processing(iter_uuids(),
                                                    hash2uuids_kvstore),
                               descriptor_index, lsh_functor,
                               report_interval,
                               use_multiprocessing, True):
        # Get original value in KV-store if not in update dict.
        if hash_int not in kv_update:
            kv_update[hash_int] = hash2uuids_kvstore.get(hash_int, set())
        kv_update[hash_int] |= {uuid}

    if kv_update:
        log.info("Updating KV store... (%d keys)" % len(kv_update))
        hash2uuids_kvstore.add_many(kv_update)

    log.info("Done")


if __name__ == '__main__':
    main()



import flask
import functools
import json
import os.path

import six
import requests

from smqtk.utils import SmqtkObject
from smqtk.utils.url import url_join


script_dir = os.path.dirname(os.path.abspath(__file__))


# noinspection PyUnusedLocal
class LoginMod (SmqtkObject, flask.Blueprint):

    def __init__(self, name, parent_app, url_prefix=None):
        """
        Initialize the login module

        :param parent_app: Parent application that is loading this using this
            module.
        :type parent_app: flask.Flask

        """
        super(LoginMod, self).__init__(
            name, __name__,
            template_folder=os.path.join(script_dir, 'templates'),
            url_prefix=url_prefix
        )

        # Pull in users configuration file from etc directory
        users_config_file = os.path.join(script_dir, 'users.json')
        if not os.path.isfile(users_config_file):
            raise RuntimeError("Couldn't find expected users config file -> %s"
                               % users_config_file)
        with open(users_config_file) as infile:
            self.__users = json.load(infile)

        #
        # Routing
        #
        @self.route('/login', methods=['get'])
        def login():
            # Render the login page, then continue on to a previously requested
            # page, or the home page.
            # noinspection PyUnresolvedReferences
            return flask.render_template(
                'login.html',
                next=flask.request.args.get('next', '/'),
                username=flask.request.args.get('username', ''),
            )

        @self.route('/login.passwd', methods=['post'])
        def login_password():
            """
            Log-in processing method called when submitting form displayed by
            the login() method above.

            There will always be a 'next' value defined in the form (see above
            method).

            """
            userid = flask.request.form['login']
            next_page = flask.request.form['next']

            if userid in self.__users:
                # Load user
                user = self.__users[userid]
                self._log.debug("User info selected: %s", user)

                if flask.request.form['passwd'] != user['passwd']:
                    flask.flash("Authentication error for user id: %s" % userid,
                                "error")
                    return flask.redirect("/login?next=%s&username=%s"
                                          % (next_page, userid))

                flask.flash("Loading user: %s" % userid, "success")
                self._login_user(userid, user)
                return flask.redirect(next_page)
            else:
                flask.flash("Unknown user: %s" % userid, 'error')
                return flask.redirect("/login?next=%s" % next_page)

        @self.route('/logout')
        def logout():
            """
            Logout the current user
            """
            if 'user' in flask.session:
                del flask.session['user']
            else:
                flask.flash("No user currently logged in!", "error")
            return flask.redirect("/")

    #
    # Utility methods
    #

    @staticmethod
    def _login_user(userid, user_info):
        """
        "log-in" the user in the current session. This adds the name and role
        list to the session. Only one user logged in at a time.

        :param userid: String ID of the user
        :type userid: str

        :param user_info: The user dictionary as recorded in our users.json
            config file.
        :type user_info: dict[str, str or list of str]

        """
        flask.session['user'] = {
            'id': userid,
            'fullname': user_info['fullname'],
            'roles': user_info['roles']
        }

    @staticmethod
    def login_required(f):
        """
        Decorator for URLs that require login.

        Girder aware. Login-required URLs can be passed:

            "girder_token"
            "girder_origin"
            "girder_apiRoot"

        argument,  We will check the given token with the origin URL and API
        root.

        :param f: Function to be wrapped.

        """
        log = LoginMod.get_logger()

        @functools.wraps(f)
        def deco(*args, **kwds):
            # Combine to handle both GET arguments and form data
            c = {}
            c.update(dict(six.iteritems(flask.request.form)))
            c.update(dict(six.iteritems(flask.request.args)))
            log.debug("Combined arguments: %s", c)

            if not {'girder_token', 'girder_origin', 'girder_apiRoot'} \
                    .difference(c.keys()):
                g_token = c['girder_token']
                log.debug("G-token: %s", g_token)
                g_origin = c['girder_origin']
                log.debug("G-origin: %s", g_origin)
                g_apiRoot = c['girder_apiRoot']
                log.debug("G-apiRoot: %s", g_apiRoot)

                g_api_header = {'Girder-Token': g_token}

                # Attempt getting current user using supplied token. If it
                # succeeds and matches the given user, log them in here.

                # Get user ID from token
                log.debug('Getting Girder current token info')
                r = requests.get(url_join(g_origin, g_apiRoot, 'token/current'),
                                 headers=g_api_header)
                if r.json() is None:
                    flask.flash("Invalid Girder token credentials token",
                                'error')
                    return flask.redirect(flask.url_for("login.login")
                                          + "?next=" + flask.request.url)
                else:
                    g_userId = r.json()['userId']

                    # Get user label and name from ID
                    log.debug("Getting user info from ID")
                    r = requests.get(url_join(g_origin, g_apiRoot, 'user',
                                              g_userId),
                                     headers=g_api_header)
                    user_label = r.json()['login']
                    user_fullname = ' '.join([r.json()['firstName'],
                                              r.json()['lastName']])

                    log.debug("Logging user '%s' in", user_fullname)
                    # Arbitrarily log the user in as a guest
                    LoginMod._login_user(user_label, {
                        'fullname': user_fullname,
                        'roles': ['guest'],
                    })

            if 'user' not in flask.session:
                flask.flash("Login required!", 'error')
                return flask.redirect(flask.url_for("login.login")
                                      + "?next=" + flask.request.url)
            else:
                # TODO: Check that user has permission, else redirect
                return f(*args, **kwds)
        return deco



import mock
import pickle
import unittest

from smqtk.exceptions import ReadOnlyError
from smqtk.representation.key_value.postgres import PostgresKeyValueStore


if PostgresKeyValueStore.is_usable():

    class TestPostgresKeyValueStore (unittest.TestCase):

        @mock.patch('smqtk.utils.postgres.get_connection_pool')
        def test_remove_readonly(self, m_gcp):
            """ Test that we cannot remove from readonly instance. """
            s = PostgresKeyValueStore(read_only=True)

            self.assertRaises(
                ReadOnlyError,
                s.remove, 0
            )

        @mock.patch('smqtk.utils.postgres.get_connection_pool')
        def test_remove_invalid_key(self, m_gcp):
            """
            Simulate an missing key and that it should result in a thrown
            KeyError
            """
            s = PostgresKeyValueStore()

            # Pretend this store contains nothing.
            s.has = mock.Mock(return_value=False)

            self.assertRaises(
                KeyError,
                s.remove, 0
            )

        @mock.patch('smqtk.utils.postgres.get_connection_pool')
        def test_remove(self, m_gcp):
            """
            Simulate removing a value from the store. Checking executions on
            the mock cursor.
            """
            expected_key = 'test_remove_key'
            expected_key_bytea = bytes(
                PostgresKeyValueStore._py_to_bin(expected_key))

            # Cut out create table calls.
            s = PostgresKeyValueStore(create_table=False)
            # Pretend key exists in index.
            s.has = mock.Mock(return_value=True)

            # Cursor is created via a context (i.e. __enter__()
            #: :type: mock.Mock
            mock_execute = s._psql_helper.get_psql_connection().cursor()\
                            .__enter__().execute

            s.remove(expected_key)

            # Call to ensure table and call to remove.
            mock_execute.assert_called_once()
            # Call should have been with provided key as converted to postgres
            # bytea type.
            self.assertRegexpMatches(mock_execute.call_args[0][0],
                                     "DELETE FROM .+ WHERE .+ LIKE .+")
            self.assertEqual(set(mock_execute.call_args[0][1].keys()),
                             {'key_like'})
            self.assertEqual(bytes(mock_execute.call_args[0][1]['key_like']),
                             expected_key_bytea)

        @mock.patch('smqtk.utils.postgres.get_connection_pool')
        def test_remove_many_readonly(self, m_gcp):
            """
            Test failure to remove from a readonly instance.
            """
            s = PostgresKeyValueStore(read_only=True)
            self.assertRaises(
                ReadOnlyError,
                s.remove_many, [0, 1]
            )

        @mock.patch('smqtk.utils.postgres.get_connection_pool')
        def test_remove_many_invalid_keys(self, m_gcp):
            """
            Test failure when one or more provided keys are not present in
            store.
            """
            s = PostgresKeyValueStore(create_table=False)

            # Simulate the batch execute returning nothing.  This simulates no
            # rows being found by the first call to the method when checking
            # for key presence in table.
            s._check_contained_keys = mock.Mock(return_value={0, 1})
            PY2_SET_KEY_ERROR_RE = "set\(\[(?:0|1), (?:0|1)\]\)"
            PY3_SET_KEY_ERROR_RE = "{(?:0|1), (?:0|1)}"
            self.assertRaisesRegexp(
                KeyError, '^(?:{}|{})$'.format(PY2_SET_KEY_ERROR_RE,
                                               PY3_SET_KEY_ERROR_RE),
                s.remove_many, [0, 1]
            )

            # Simulate only one of the keys existing in the table.
            s._check_contained_keys = mock.Mock(return_value={1})
            self.assertRaisesRegexp(
                KeyError, '^1$',
                s.remove_many, [0, 1]
            )
            s._check_contained_keys = mock.Mock(return_value={0})
            self.assertRaisesRegexp(
                KeyError, '^0$',
                s.remove_many, [0, 1]
            )

        @mock.patch('smqtk.utils.postgres.get_connection_pool')
        @mock.patch('smqtk.representation.key_value.postgres.psycopg2.extras.execute_batch')
        def test_remove_many(self, m_psqlExecBatch, m_gcp):
            """
            Test expected calls to psql cursor during normal operation.
            """
            expected_key_1 = 'test_remove_many_key_1'
            exp_key_1_bytea = PostgresKeyValueStore._py_to_bin(expected_key_1)
            expected_key_2 = 'test_remove_many_key_2'
            exp_key_2_bytea = PostgresKeyValueStore._py_to_bin(expected_key_2)

            # Skip table creation calls.
            s = PostgresKeyValueStore(create_table=False)

            # Cursor is created via a context (i.e. __enter__()
            #: :type: mock.Mock
            mock_cursor = s._psql_helper.get_psql_connection().cursor() \
                .__enter__()
            # Make the cursor iterate keys in order to pass key check.
            mock_cursor.__iter__.return_value = [[pickle.dumps(expected_key_1)],
                                                 [pickle.dumps(expected_key_2)]]
            #: :type: mock.Mock
            mock_execute = mock_cursor.execute

            s.remove_many([expected_key_1, expected_key_2])

            # As a result of this call, we expect:
            # - ``cursor.execute`` should have been called once when checking
            #   for key presence in db (query < batch size)
            # - ``psycopg2.extras.execute_batch`` should have been called once
            #   when deleting key-value pairs in db (query < batch size)
            #
            # We back to break up the argument equality check of recorded mock
            # funciton call arguments due to ``psycopg2.Binary`` instances not
            # being comparable.

            mock_execute.assert_called_once()
            m_psqlExecBatch.assert_called_once()

            # Confirm call arguments to ``cursor.execute``
            expected_has_q = "SELECT key FROM data_set WHERE key IN " \
                             "%(key_tuple)s"
            mock_execute_call_args = mock_execute.call_args[0]
            self.assertEqual(mock_execute_call_args[0], expected_has_q)
            self.assertEqual(set(mock_execute_call_args[1].keys()), {'key_tuple'})
            self.assertEqual(
                [bytes(k) for k in mock_execute_call_args[1]['key_tuple']],
                [bytes(k) for k in [exp_key_1_bytea, exp_key_2_bytea]]
            )

            # Confirm call arguments to ``psycopg2.extras.execute_batch``
            expected_del_q = "DELETE FROM data_set WHERE key LIKE %(key_like)s"
            expected_del_vals = [{'key_like': exp_key_1_bytea},
                                 {'key_like': exp_key_2_bytea}]
            psqlExecBatch_call_args = m_psqlExecBatch.call_args[0]
            self.assertEqual(psqlExecBatch_call_args[0], mock_cursor)
            self.assertEqual(psqlExecBatch_call_args[1], expected_del_q)
            # 3rd argument is a list of dictionaries for 'key_like' replacements
            self.assertEqual(len(psqlExecBatch_call_args[2]), 2)
            self.assertListEqual(
                [bytes(d['key_like']) for d in psqlExecBatch_call_args[2]],
                [bytes(exp_key_1_bytea), bytes(exp_key_2_bytea)]
            )



#!/usr/bin/env python
from __future__ import print_function
import os
import re
import setuptools


PYTHON_SRC = 'python'
PYTHON_FILE_RE = re.compile('.*\.(?:py[co]?)$')


# TODO: Auto-append dev version based on how forward from latest release
#       Basically a simpler version of what setuptools_scm does but without
#       the added cruft and bringing the ENTIRE git repo in with the dist
#       See: https://github.com/pypa/setuptools_scm/blob/master/setuptools_scm/version.py
#       Would need to know number of commits ahead from last version tag.
with open("VERSION") as f:
    version = f.read().strip()


with open('README.md') as f:
    long_description = f.read()


def find_package_datafiles(package_dir):
    """ Return a list of non-python files in package source tree

    File paths are relative to the top of the package directory provided.
    """
    # TODO: Add exclusion list/glob/regex parameter if necessary.
    non_python = set()
    for dirpath, _, fnames in os.walk(package_dir):
        non_python.update([os.path.relpath(os.path.join(dirpath, fp),
                                           package_dir)
                           for fp in fnames
                           # Things that are NOT python files
                           if PYTHON_FILE_RE.match(fp) is None])
    return list(non_python)


def list_directory_files(dirpath, exclude_dirs=(), exclude_files=()):
    """
    List files and their parent directories in the format required for the
    ``setup`` function ``data_files`` parameter:

        ...
        data_files=[
            ('dir', 'root-relative-file-path'),
            ...
        ],
        ...

    This function is intended to effectively install a directory located in the
    source root as is (e.g. the ``etc`` directory).

    :param dirpath: Base directory to start with. The directory paths returned
        start with this directory.
    :param exclude_dirs: sequence if directory paths (starting from ``dirpath``)
        that should not be included. For example, we don't want the `bin/memex'
        directory to be installed, when gathering data files for `bin`, we call
        this function like:

            list_directory_files('bin', ['bin/memex'])
    :param exclude_files: File names to ignore in directories traversed.

    """
    exclude_dirs = set(ed.strip(' /') for ed in exclude_dirs)
    exclude_files = set(ef.strip() for ef in exclude_files)
    file_paths = []
    for dirpath, dnames, fnames in os.walk(dirpath):
        # Filter out directories to be excluded
        for dn in dnames:
            if os.path.join(dirpath, dn) in exclude_dirs:
                print("skipping:", os.path.join(dirpath, dn))
                del dnames[dnames.index(dn)]
        # filter out excluded files
        fnames = set(fnames).difference(exclude_files)
        # collect directory to file paths reference
        file_paths.append(
            (dirpath, [os.path.join(dirpath, fn) for fn in fnames])
        )
    return file_paths


################################################################################


setuptools.setup(
    name='smqtk',
    version=version,
    description='Python toolkit for pluggable algorithms and data structures '
                'for multimedia-based machine learning',
    long_description=long_description,
    author='Kitware, Inc.',
    author_email='smqtk-developers@kitware.com',
    url='https://github.com/Kitware/SMQTK',
    license='BSD 3-Clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Recognition'
    ],
    platforms=[
        'Linux',
        'Max OS-X',
        'Unix',
        # 'Windows',  # Not tested yet
    ],

    package_dir={'': PYTHON_SRC},
    packages=setuptools.find_packages(PYTHON_SRC),
    package_data={
        'smqtk': find_package_datafiles(os.path.join(PYTHON_SRC, 'smqtk'))
    },
    data_files=list_directory_files('etc'),

    setup_requires=[
        'setuptools',
    ],
    install_requires=[
        'flask',
        'flask-basicauth',
        'flask-login',
        'Flask-Cors',
        'imageio',
        'jinja2',
        'matplotlib',
        'numpy',
        'pillow',
        'pymongo',
        'requests',
        'scikit-learn',
        'scipy',
        'six',
    ],
    extras_require={
        # Various optional dependencies for plugins
        'docs': [
            'livereload',
            'mock',
            'sphinx',
            'sphinx-argparse',
            'sphinx-prompt',
            'sphinx_rtd_theme',
        ],
        'caffe': [
            'protobuf',
            'scikit-image',
        ],
        'magic': [
            'file-magic',
        ],
        'flann': [
            'pyflann>=1.8.4',
        ],
        'libmagic': [
            'file-magic',
        ],
        'postgres': [
            'psycopg2',
        ],
        'solr': [
            'solrpy',
        ],
        'girder': [
            'girder-client',
        ]
    },
    tests_require=[
        'coverage',
        'mock',
        'pytest',
        'pytest-cov',
    ],

    # See entry_points/console_scripts as the preferred method for publishing
    #   executable scripts. Might have redesign how scripts are done if that is
    #   to be used...
    # TODO: Rename camel-case scripts to ``smqtk-...`` format without camel-case
    entry_points={
        'console_scripts': [
            'classifier_kfold_validation = \
                smqtk.bin.classifier_kfold_validation\
                :classifier_kfold_validation',
            'classifier_model_validation = \
                smqtk.bin.classifier_model_validation:main',
            'smqtk-classify-files = smqtk.bin.classifyFiles:main ',
            'compute_classifications = smqtk.bin.compute_classifications:main',
            'compute_hash_codes = smqtk.bin.compute_hash_codes:main',
            'compute_many_descriptors = \
                smqtk.bin.compute_many_descriptors:main',
            'smqtk-compute-descriptor = smqtk.bin.computeDescriptor:main',
            'smqtk-create-file-ingest = smqtk.bin.createFileIngest:main',
            'smqtk-create-girder-ingest = smqtk.bin.createGirderIngest:main',
            'descriptors_to_svmtrain = \
                smqtk.bin.descriptors_to_svmtrainfile:main',
            'generate_image_transform = \
                smqtk.bin.generate_image_transform:main',
            'iqr_app_model_generation = \
                smqtk.bin.iqr_app_model_generation:main',
            'iqrTrainClassifier = smqtk.bin.iqrTrainClassifier:main',
            'make_balltree = smqtk.bin.make_balltree:main',
            'minibatch_kmeans_clusters = \
                smqtk.bin.minibatch_kmeans_clusters:main',
            'smqtk-remove-old-files = smqtk.bin.removeOldFiles:main',
            'smqtk-proxy-manager-server = smqtk.bin.proxyManagerServer:main',
            'runApplication = smqtk.bin.runApplication:main',
            'smqtk-summarize-plugins = smqtk.bin.summarizePlugins:main',
            'train_itq = smqtk.bin.train_itq:main',
            'smqtk-make-train-test-sets = smqtk.bin.make_train_test_sets:main',
            'smqtk-nearest-neighbors = smqtk.bin.nearest_neighbors:main',
            'smqtk-check-images = smqtk.bin.check_images:main'
        ]
    }
)




import flask
import multiprocessing
import os
import tempfile

from six.moves import cStringIO as StringIO

from smqtk.utils import SmqtkObject
from smqtk.utils import file_utils


script_dir = os.path.dirname(os.path.abspath(__file__))


class FileUploadMod (SmqtkObject, flask.Blueprint):
    """
    Flask blueprint for file uploading.
    """

    def __init__(self, name, parent_app, working_directory, url_prefix=None):
        """
        Initialize uploading module

        :param parent_app: Parent Flask app
        :type parent_app: smqtk.Web.search_app.base_app.search_app

        :param working_directory: Directory for temporary file storage during
            upload up to the time a user takes control of the file.
        :type working_directory: str

        """
        super(FileUploadMod, self).__init__(
            name, __name__,
            static_folder=os.path.join(script_dir, 'static'),
            url_prefix=url_prefix
        )
        # TODO: Thread safety

        self.parent_app = parent_app
        self.working_dir = working_directory

        # TODO: Move chunk storage to database for APACHE multiprocessing
        # File chunk aggregation
        #   Top level key is the file ID of the upload. The dictionary
        #   underneath that is the index ID'd chunks. When all chunks are
        #   present, the file is written and the entry in this map is removed.
        #: :type: dict of (str, dict of (int, StringIO))
        self._file_chunks = {}
        # Lock per file ID so as to not collide when uploading multiple chunks
        #: :type: dict of (str, RLock)
        self._fid_locks = {}

        # FileID to temporary path that a completed file is located at.
        self._completed_files = {}

        #
        # Routing
        #

        @self.route('/upload_chunk', methods=["POST"])
        @self.parent_app.module_login.login_required
        def upload_file():
            """
            Handle arbitrary file upload to OS temporary file storage, recording
            file upload completions.

            """
            form = flask.request.form
            self._log.debug("POST form contents: %s" % str(flask.request.form))

            fid = form['flowIdentifier']
            current_chunk = int(form['flowChunkNumber'])
            total_chunks = int(form['flowTotalChunks'])
            filename = form['flowFilename']

            #: :type: FileStorage
            chunk_data = flask.request.files['file']

            with self._fid_locks.setdefault(fid, multiprocessing.RLock()):
                # Create new entry in chunk map / add to existing entry
                # - Need to explicitly copy the buffered data as the file object
                #   closes between chunk messages.
                self._file_chunks.setdefault(fid, {})[current_chunk] \
                    = StringIO(chunk_data.read())
                message = "Uploaded chunk #%d of %d for file '%s'" \
                    % (current_chunk, total_chunks, filename)

                if total_chunks == len(self._file_chunks[fid]):
                    self._log.debug("[%s] Final chunk uploaded",
                                    filename+"::"+fid)
                    # have all chucks in memory now
                    try:
                        # Combine chunks into single file
                        file_ext = os.path.splitext(filename)[1]
                        file_saved_path = self._write_file_chunks(
                            self._file_chunks[fid], file_ext
                        )
                        self._log.debug("[%s] saved from chunks: %s",
                                        filename+"::"+fid, file_saved_path)
                        # now in file, free up dict memory

                        self._completed_files[fid] = file_saved_path
                        message = "[%s] Completed upload" % (filename+"::"+fid)

                    except IOError as ex:
                        self._log.debug("[%s] Failed to write combined chunks",
                                        filename+"::"+fid)
                        message = "Failed to write out combined chunks for " \
                                  "file %s: %s" % (filename, str(ex))
                        raise RuntimeError(message)

                    except NotImplementedError as ex:
                        message = "Encountered non-implemented code path: %s" \
                                  % str(ex)
                        raise RuntimeError(message)

                    finally:
                        # remove chunk map entries
                        del self._file_chunks[fid]
                        del self._fid_locks[fid]

            # Flow only displays return as a string, so just returning the
            # message component.
            return message

        @self.route("/completed_uploads")
        @self.parent_app.module_login.login_required
        def completed_uploads():
            return flask.jsonify(self._completed_files)

    def upload_post_url(self):
        """
        :return: The url string to give to the JS upload zone for POSTing file
            chunks.
        :rtype: str
        """
        return self.url_prefix + '/upload_chunk'

    def get_path_for_id(self, file_unique_id):
        """
        Get the path to the temp file that was uploaded.

        It is the user's responsibility to remove this file when it is done
        being used, or move it else where.

        :param file_unique_id: Unique ID of the uploaded file
        :type file_unique_id: str

        :return: The path to the complete uploaded file.

        """
        return self._completed_files[file_unique_id]

    def clear_completed(self, file_unique_id):
        """
        Clear the completed file entry in our cache. This should be called after
        taking responsibility for an uploaded file.

        This does NOT delete the file.

        :raises KeyError: If the given unique ID does not correspond to an
            entry in our completed cache.

        :param file_unique_id: Unique ID of an uploaded file to clear from the
            completed cache.
        :type file_unique_id: str

        """
        del self._completed_files[file_unique_id]

    # noinspection PyMethodMayBeStatic
    def _write_file_chunks(self, chunk_map, file_extension=''):
        """
        Given a mapping of chunks, write their contents to a temporary file,
        returning the path to that file.

        Returned file path should be manually removed by the user.

        :param chunk_map: Mapping of integer index to file-like chunk
        :type chunk_map: dict of (int, StringIO)
        :param file_extension: String extension to suffix the temporary file
            with
        :type file_extension: str

        :raises OSError: OS problems creating temporary file or writing it out.

        :return: Path to temporary combined file
        :rtype: str

        """
        # Make sure write dir exists...
        if not os.path.isdir(self.working_dir):
            file_utils.safe_create_dir(self.working_dir)
        tmp_fd, tmp_path = tempfile.mkstemp(file_extension,
                                            dir=self.working_dir)
        self._log.debug("Combining chunks into temporary file: %s", tmp_path)
        tmp_file = open(tmp_path, 'wb')
        for idx, chunk in sorted(chunk_map.items(), key=lambda p: p[0]):
            data = chunk.read()
            tmp_file.write(data)
        tmp_file.close()
        return tmp_path
